unit ScanMan;

interface

uses SysUtils, Classes;

const
  ScanManDLL = 'ScanMan.dll';

  { Parameter types }

  SM_TYPE_VOID   = 0;
  SM_TYPE_BYTE   = 1;
  SM_TYPE_SBYTE  = 2;
  SM_TYPE_WORD   = 3;
  SM_TYPE_SHORT  = 4;
  SM_TYPE_DWORD  = 5;
  SM_TYPE_INT    = 6;
  SM_TYPE_BOOL   = 10;
  SM_TYPE_CHAR   = 20;
  SM_TYPE_STRING = 100;

  { Events }

  SM_EVENT_PLUGGED   = 1;
  SM_EVENT_UNPLUGGED = 2;
  SM_EVENT_ERROR     = 3;

  { States }

  SM_STATE_FINGER_DOWN = 1;
  SM_STATE_FINGER_UP   = 2;

  { Error codes }

  // General
  SME_OK                  = 0;
  SME_FAILED              = -1;
  SME_OUT_OF_MEMORY       = -2;
  SME_NOT_INITIALIZED     = -3;
  SME_ARGUMENT_NULL       = -4;
  SME_INVALID_ARGUMENT    = -5;
  // Parameters
  SME_INVALID_PARAMETER   = -10;
  SME_PARAMETER_READ_ONLY = -11;
  // Capturing
  SME_NOT_CAPTURING       = -100;
  SME_ALREADY_CAPTURING   = -101;

  { Parameters }

  // General
  SMP_TYPE         = 0; // Get | Pass parameter identifier ((VOID*)SMP_XXX) as value | Returns one of SM_TYPE_XXX constants
  SMP_NAME         = 10; // Get | String
  SMP_VERSION_HIGH = 11; // Get | DWord
  SMP_VERSION_LOW  = 12; // Get | DWord
  SMP_COPYRIGHT    = 13; // Get | String

type

  { Callbacks }

  TSMMonitorProc = procedure(const ScannerId: PAnsiChar; Event: LongWord; Param: Pointer); stdcall;

  TSMImageProc = procedure(const ScannerId: PAnsiChar; Width, Height: Integer;
    const Image: PByte; Resolution: Integer; Param: Pointer); stdcall;
  TSMStateProc = procedure(const ScannerId: PAnsiChar; State: LongWord; Param: Pointer); stdcall;

{ ScanMan library functions }

// Initialization
function SMInitialize: Integer; stdcall;
function SMFinalize: Integer; stdcall;

// Parameters
function SMGetParameter(Reserved: Integer; Parameter: Integer; Value: Pointer): Integer; stdcall;
function SMSetParameter(Reserved: Integer; Parameter: Integer; Value: LongWord): Integer; stdcall;

// Scanner enumeration
function SMGetScannerCount: Integer; stdcall;
function SMGetScannerId(Index: Integer; Id: PAnsiChar; Len: Integer): Integer; stdcall; overload;

// Scanner monitoring
function SMSetMonitor(MonitorProc: TSMMonitorProc; Param: Pointer): Integer; stdcall;
function SMRemoveMonitor: Integer; stdcall;

// Capturing
function SMStartCapturing(ScannerId: PAnsiChar; ImageProc: TSMImageProc;
  StateProc: TSMStateProc; Param: Pointer): Integer; stdcall; overload;
function SMStopCapturing(ScannerId: PAnsiChar): Integer; stdcall; overload;

{ Error and result functions }

function SMSucceeded(Res: Integer): Boolean;
function SMFailed(Res: Integer): Boolean;

function SMErrorToString(Err: Integer): string;
function SMResultToString(Res: Integer): string;

procedure SMCheckResult(Res: Integer);

{ Parameters functions }

function SMParameterOfType(Reserved: Integer; Parameter, ParameterType: Integer): Boolean;
procedure SMCheckParameterType(Reserved: Integer; Parameter, ParameterType: Integer);

// Get
function SMGetDWordParameter(Reserved: Integer; Parameter: Integer): LongWord;
function SMGetStringParameter(Reserved: Integer; Parameter: Integer): string;

// Get general
function SMGetParameterType(Reserved: Integer; Parameter: Integer): Integer;
function SMGetName(Reserved: Integer = 0): string;
function SMGetVersionHigh(Reserved: Integer = 0): LongWord;
function SMGetVersionLow(Reserved: Integer = 0): LongWord;
function SMGetCopyright(Reserved: Integer = 0): string;

{ Scanner enumeration functions }

function SMGetScannerId(Index: Integer): string; overload;
procedure SMGetScannerIds(Ids: TStrings);


{ Capturing functions }

function SMStartCapturing(ScannerId: string; ImageProc: TSMImageProc;
  StateProc: TSMStateProc; Param: Pointer): Integer; overload;
function SMStopCapturing(ScannerId: string): Integer; overload;
function SMStopCapturing: Integer; overload;


{ Event and state functions }

function SMEventToString(Event: LongWord; ScannerId: string = ''): string;
function SMStateToString(State: LongWord; ScannerId: string = ''): string;

implementation

{ ScanMan library functions }

function SMInitialize: Integer; external ScanManDLL;
function SMFinalize: Integer; external ScanManDLL;
function SMGetParameter(Reserved: Integer; Parameter: Integer; Value: Pointer): Integer; external ScanManDLL;
function SMSetParameter(Reserved: Integer; Parameter: Integer; Value: LongWord): Integer; external ScanManDLL;
function SMGetScannerCount: Integer; external ScanManDLL;
function SMGetScannerId(Index: Integer; Id: PAnsiChar; Len: Integer): Integer; external ScanManDLL;
function SMSetMonitor(MonitorProc: TSMMonitorProc; Param: Pointer): Integer; external ScanManDLL;
function SMRemoveMonitor: Integer; external ScanManDLL;
function SMStartCapturing(ScannerId: PAnsiChar; ImageProc: TSMImageProc;
  StateProc: TSMStateProc; Param: Pointer): Integer; external ScanManDLL;
function SMStopCapturing(ScannerId: PAnsiChar): Integer; external ScanManDLL;

{ Error and result functions }

function SMSucceeded(Res: Integer): Boolean;
begin
  Result := Res >= 0;
end;

function SMFailed(Res: Integer): Boolean;
begin
  Result := Res < 0;
end;

function SMErrorToString(Err: Integer): string;
begin
  if Err >= 0 then Result := EmptyStr
  else
    case Err of
      SME_FAILED: Result := 'Failed';
      SME_OUT_OF_MEMORY: Result := 'Out of memory';
      SME_NOT_INITIALIZED: Result := 'Not initialized';
      SME_ARGUMENT_NULL: Result := 'Argument null';
      SME_INVALID_ARGUMENT: Result := 'Invalid argument';
      SME_INVALID_PARAMETER: Result := 'Invalid parameter';
      SME_PARAMETER_READ_ONLY: Result := 'Parameter is read only';
      SME_ALREADY_CAPTURING: Result := 'Already capturing';
      SME_NOT_CAPTURING: Result := 'Not capturing';
    else Result := IntToStr(Err);
    end;
end;

function SMResultToString(Res: Integer): string;
begin
  if SMFailed(Res) then Result := SMErrorToString(Res)
  else if Res = SME_OK then Result := 'OK' else Result := Format('OK (%d)', [Res]);
end;

procedure SMCheckResult(Res: Integer);
begin
  if SMFailed(Res) then
    raise Exception.CreateFmt('ScanMan error:'#13#9'Code = %d'#13#9'Description = %s',
      [Res, SMErrorToString(Res)]);
end;

{ Parameters functions }

function SMParameterOfType(Reserved: Integer; Parameter, ParameterType: Integer): Boolean;
begin
  Result := SMGetParameterType(Reserved, Parameter) = ParameterType;
end;

procedure SMCheckParameterType(Reserved: Integer; Parameter, ParameterType: Integer);
begin
  if not SMParameterOfType(Reserved, Parameter, ParameterType) then
    raise Exception.Create('ScanMan parameter is of wrong type');
end;

// Get

function SMGetDWordParameter(Reserved: Integer; Parameter: Integer): LongWord;
begin
  SMCheckParameterType(Reserved, Parameter, SM_TYPE_DWORD);
  SMCheckResult(SMGetParameter(Reserved, Parameter, @Result));
end;

function SMGetStringParameter(Reserved: Integer; Parameter: Integer): string;
var
  L: Integer;
  Value: AnsiString;
begin
  SMCheckParameterType(Reserved, Parameter, SM_TYPE_STRING);
  L := SMGetParameter(Reserved, Parameter, nil);
  SMCheckResult(L);
  SetLength(Value, L);
  SMCheckResult(SMGetParameter(Reserved, Parameter, PAnsiChar(Value)));
  Result := Value;
end;

// Get general

function SMGetParameterType(Reserved: Integer; Parameter: Integer): Integer;
begin
  Result := SMGetParameter(Reserved, SMP_TYPE, Pointer(Parameter));
  SMCheckResult(Result);
end;

function SMGetName(Reserved: Integer = 0): string;
begin
  Result := SMGetStringParameter(Reserved, SMP_NAME);
end;

function SMGetVersionHigh(Reserved: Integer = 0): LongWord;
begin
  Result := SMGetDWordParameter(Reserved, SMP_VERSION_HIGH);
end;

function SMGetVersionLow(Reserved: Integer = 0): LongWord;
begin
  Result := SMGetDWordParameter(Reserved, SMP_VERSION_LOW);
end;

function SMGetCopyright(Reserved: Integer = 0): string;
begin
  Result := SMGetStringParameter(Reserved, SMP_COPYRIGHT);
end;

{ Scanner enumeration functions }

function SMGetScannerId(Index: Integer): string;
var
  L: Integer;
  Id: AnsiString;
begin
  L := SMGetScannerId(Index, nil, 0);
  if SMSucceeded(L) then
  begin
    SetLength(Id, L - 1);
    L := SMGetScannerId(Index, PAnsiChar(Id), L);
    if SMSucceeded(L) then
    begin
      Result := Id;
      Exit;
    end;
  end;
  raise Exception.Create('Can not retrieve scanner id');
end;

procedure SMGetScannerIds(Ids: TStrings);
var
  C, I: Integer;
begin
  Ids.Clear;
  C := SMGetScannerCount;
  if SMSucceeded(C) then
  begin
    for I := 0 to C - 1 do
      try
        Ids.Add(SMGetScannerId(I));
      except
      end;
  end;
end;

{ Capturing functions }

function SMStartCapturing(ScannerId: string; ImageProc: TSMImageProc;
  StateProc: TSMStateProc; Param: Pointer): Integer;
var
  Id: AnsiString;
begin
  Id := ScannerId;
  Result := SMStartCapturing(PAnsiChar(Id), ImageProc, StateProc, Param);
end;

function SMStopCapturing(ScannerId: string): Integer;
var
  Id: AnsiString;
begin
  Id := ScannerId;
  Result := SMStopCapturing(PAnsiChar(Id));
end;

function SMStopCapturing: Integer;
begin
  Result := SMStopCapturing(nil);
end;

{ Event and state functions }

function SMEventToString(Event: LongWord; ScannerId: string = ''): string;
begin
  case Event of
    SM_EVENT_PLUGGED: Result := 'Plugged';
    SM_EVENT_UNPLUGGED: Result := 'Unplugged';
    SM_EVENT_ERROR: Result := 'Error';
  else
    Result := IntToStr(Event);
  end;
  if ScannerId <> EmptyStr then Result := Result + ': ' + ScannerId;
end;

function SMStateToString(State: LongWord; ScannerId: string = ''): string;
begin
  case State of
    SM_STATE_FINGER_DOWN:
      if ScannerId = EmptyStr then Result := 'Finger down'
      else Result := 'Finger placed onto ' + ScannerId;
    SM_STATE_FINGER_UP:
      if ScannerId = EmptyStr then Result := 'Finger up'
      else Result := 'Finger removed from ' + ScannerId;
  else
    Result := IntToStr(State);
    if ScannerId <> EmptyStr then Result := Result + ' on ' + ScannerId;
  end;
end;

end.
