#include "StdAfx.h"
#include "afxwin.h"
#include "VFDDatabase.h"
#include "VFDUtils.h"
#include <exception>

VFDDatabase::VFDDatabase()
	: connected(false), opened(false), recordsetName("Fingerprints"),
	connection(NULL), recordset(NULL), recordBinding(NULL)
{
	for(int g = 0; g < 256; g++)
	{
		items[g] = NULL;
		capacity[g] = 0;
		count[g] = 0;
	}
	allItems = NULL;
	allCount = 0;
	allCapacity = 0;
}

VFDDatabase::~VFDDatabase()
{
	Close();
}

int VFDDatabase::ID(int index)
{
	CheckIndex(index);
	return allItems[index]->id;
}

const BYTE * VFDDatabase::Features(int index)
{
	CheckIndex(index);
	return allItems[index]->features;
}

DWORD VFDDatabase::FeaturesSize(int index)
{
	CheckIndex(index);
	return allItems[index]->featuresSize;
}

CString VFDDatabase::FingerID(int index)
{
	CheckIndex(index);
	return allItems[index]->fingerId;
}

#define VFDBC(c, v) c <= v ? v :
#define VFDBCapacity(c) (c <= 2 ? c : VFDBC(c, 4) VFDBC(c, 8) VFDBC(c, 16) VFDBC(c, 32) VFDBC(c, 64) VFDBC(c, 128) VFDBC(c, 256) VFDBC(c, 512) (c / 1024 + (c % 1024 ? 1 : 0) * 1024))
#define VFDBAllCapacity(c) (c <= 2 ? c : VFDBC(c, 4) VFDBC(c, 8) VFDBC(c, 16) VFDBC(c, 32) VFDBC(c, 64) VFDBC(c, 128) VFDBC(c, 256) VFDBC(c, 512) \
	VFDBC(c, 1024) VFDBC(c, 2048) VFDBC(c, 4096) VFDBC(c, 8192) VFDBC(c, 16384) VFDBC(c, 32768) VFDBC(c, 65536) (c / 131072 + (c % 131072 ? 1 : 0) * 131072))

#define VFDBICD(value) cp == value ? value :
#define VFDBIncCapacityDelta(cp) (cp < 2 ? 1 : VFDBICD(2) VFDBICD(4) VFDBICD(8) VFDBICD(16) VFDBICD(32) VFDBICD(64) VFDBICD(128) VFDBICD(256) VFDBICD(512) 1024)
#define VFDBIncAllCapacityDelta(cp) (cp < 2 ? 1 : VFDBICD(2) VFDBICD(4) VFDBICD(8) VFDBICD(16) VFDBICD(32) VFDBICD(64) VFDBICD(128) VFDBICD(256) VFDBICD(512) \
	VFDBICD(1024) VFDBICD(2048) VFDBICD(4096) VFDBICD(8192) VFDBICD(16384) VFDBICD(32768) VFDBICD(65536) 131072)

void VFDDatabase::OpenRecordset(ADODB::CursorTypeEnum cursorType, bool bind,
	ADODB::LockTypeEnum lockType, ADODB::CursorLocationEnum cursorLocation)
{
	CloseRecordset();
	recordset->CursorLocation = cursorLocation;
	recordset->Open((_bstr_t)recordsetName, _variant_t((IDispatch*)connection, true),
		(ADODB::CursorTypeEnum)cursorType, lockType, adCmdTableDirect);
	opened = true;
	if(bind)
	{
		TESTHR(recordset->QueryInterface(__uuidof(IADORecordBinding), (LPVOID*)&recordBinding));
		TESTHR(recordBinding->BindToRecordset(&record));
	}
}

void VFDDatabase::CloseRecordset()
{
	if(recordBinding)
	{
		recordBinding->Release();
		recordBinding = NULL;
	}
	if(opened)
	{
		recordset->Close();
		opened = false;
	}
}

void VFDDatabase::Open(const CString &provider, const CString &dataSource, const CString &userId, const CString &password)
{
	Close();

	CString connectionString;
	connectionString.Format(TEXT("Provider=%s;Data Source=%s;"), (LPCTSTR)provider, (LPCTSTR)dataSource);

	try
	{
		TESTHR(connection.CreateInstance(__uuidof(ADODB::Connection)));
		TESTHR(recordset.CreateInstance(__uuidof(ADODB::Recordset)));

		connection->Open(_bstr_t(connectionString), _bstr_t(userId), _bstr_t(password), adConnectUnspecified);
		connected = true;
		
		ADODB::_RecordsetPtr r = connection->Execute(_bstr_t("SELECT COUNT(*) AS RC FROM " + recordsetName), NULL, adCmdText);
		int rc = r->Fields->Item["RC"]->Value.lVal;

		allCapacity = VFDBAllCapacity(rc);
		allItems = (VFDDBItem**)malloc(sizeof(VFDDBItem*) * allCapacity);

		OpenRecordset(ADODB::adOpenForwardOnly);

		DWORD dt = GetTickCount();
		while(!recordset->ADOEOF)
		{
			if(record.idStatus != adFldOK || record.featuresStatus != adFldOK
				|| record.fingerIdStatus != adFldOK) {
				AfxMessageBox("Can not open database\nError reading data");
				AfxThrowUserException();	
			}
			AddItem(record.id, record.features, record.featuresLength,
				record.fingerId, record.fingerIdLength / sizeof(TCHAR));
			recordset->MoveNext();
		}

		OpenRecordset();
		
		this->provider = provider;
		this->dataSource = dataSource;
	}
	catch(_com_error &e)
	{
		try
		{
			CString error;
			CString errorMsg = "";
			errorMsg = e.ErrorMessage();
			error.Format("Can not open database : %s",errorMsg);
			AfxMessageBox(error);
			Close();
		} catch (CException* e) {
			CString error;
			LPTSTR errorMsg = "";
			e->GetErrorMessage(errorMsg,1);
			error.Format("Can not open database : %s",errorMsg);
			AfxMessageBox(error);
			AfxThrowUserException();
		}
	}
}

void VFDDatabase::Close()
{
	ClearItems();
	try
	{
		CloseRecordset();
		recordset = NULL;
		if(connected)
		{
			connection->Close();
			connected = false;
		}
		connection = NULL;
	}
	catch(_com_error &e)
	{	
		CString error;
		error.Format("Can not close database : %s", e.ErrorMessage());
		AfxMessageBox(error);
		AfxThrowUserException();
	}
	provider = "";
	dataSource = "";
}

void VFDDatabase::ClearItems()
{
	heap.Clear();
	for(int g = 0; g < 256; g++)
	{
		free(items[g]);
		items[g] = NULL;
		capacity[g] = 0;
		count[g] = 0;
	}
	free(allItems);
	allItems = NULL;
	allCount = 0;
	allCapacity = 0;
}

void VFDDatabase::Clear()
{
	ClearItems();
	if(opened)
	{
		try
		{
			CloseRecordset();
			CString cnnStr;
			cnnStr.Format("DELETE FROM %s", (LPCTSTR)recordsetName);
			connection->Execute(_bstr_t(cnnStr),
				NULL, adCmdText | adExecuteNoRecords);
			OpenRecordset();
		}
		catch(_com_error &e)
		{
			CString error;
			error.Format("Can not clear database : %s", e.ErrorMessage());
			AfxMessageBox(error);
			AfxThrowUserException();
		}
	}
}

void VFDDatabase::CopyString(LPTSTR dst, DWORD &dstLength, LPCTSTR src, DWORD srcLength, DWORD maxLength)
{
	if(srcLength > maxLength) srcLength = maxLength;
	dstLength = srcLength;
	memcpy(dst, src, srcLength * sizeof(CHAR));
	dst[srcLength] = TEXT('\0');
}

void VFDDatabase::CopyFingerID(LPTSTR dst, DWORD &dstLength, LPCTSTR src, DWORD srcLength)
{
	CopyString(dst, dstLength, src, srcLength, FINGER_ID_LENGTH);
}

int VFDDatabase::AddItem(int id, const BYTE * features, DWORD size,
	CString fingerId, DWORD fingerIdLength)
{
	int g = VFFeatGetG(features); VFCheckResult(g);
	int c = count[g];
	int cp = capacity[g];
	if(c == cp)
	{
		cp++;
		items[g] = (VFDDBItem**)realloc(items[g], sizeof(VFDDBItem*) * cp);
		capacity[g] = cp;
	}
	if(allCount == allCapacity)
	{
		allCapacity++;//= VFDBIncAllCapacityDelta(allCapacity);
		allItems = (VFDDBItem**)realloc(allItems, sizeof(VFDDBItem*) * allCapacity);
	}
	VFDDBItem *item = (VFDDBItem*) heap.Alloc((WORD)sizeof(VFDDBItem));
	item->index = allCount;
	item->gIndex = c;
	item->id = id;
	item->features = heap.Alloc((WORD)size);
	item->featuresSize = size;
	memcpy(item->features, features, size* sizeof(CHAR));
	CopyFingerID(item->fingerId, item->fingerIdLength, fingerId, fingerId.GetLength());//sizeof(fingerId)* sizeof(CHAR));

	items[g][c] = item;

	allItems[allCount] = item;
	count[g] = ++c;
	return allCount++;
}

void VFDDatabase::SetRecord(int index)
{
	VFDDBItem * item = allItems[index];
	memcpy(record.features, item->features, item->featuresSize);
	record.featuresLength = item->featuresSize;
	record.featuresStatus = adFldOK;
	record.fingerIdLength = sizeof(item->fingerId) * sizeof(TCHAR);
	memcpy(record.fingerId, item->fingerId, sizeof(item->fingerId) * sizeof(TCHAR));
	record.fingerId[item->fingerIdLength] = TEXT('\0');
	record.fingerIdStatus = adFldOK;
}

bool VFDDatabase::FindRecord(int id)
{
	try
	{
		recordset->MoveFirst();
		CString find;
		find.Format("ID=%d", id);
		recordset->Find(_bstr_t(find), 0, ADODB::adSearchForward);
		return (!recordset->ADOEOF);
	}
	catch(_com_error &e)
	{
		CString error;
		error.Format("Can not find record in the database : %s", e.ErrorMessage());
		AfxMessageBox(error);
		AfxThrowUserException();
		return false;
	}
}

void VFDDatabase::UpdateRecord(int index)
{
	SetRecord(index);
	this->FindRecord(this->allItems[index]->id);
	try
	{
		TESTHR(recordBinding->Update(&record));
	}
	catch(_com_error &e)
	{
		CString error;
		error.Format("Can not update record in the database : %s", e.ErrorMessage());
		AfxMessageBox(error);
		AfxThrowUserException();
	}
}

void VFDDatabase::CheckIndex(int index)
{
	if(index < 0 || index >= allCount) {
	 AfxMessageBox("Database record index out of bounds");
	 AfxThrowUserException();
	}
}

void VFDDatabase::SetFeatures(int index, const BYTE * features, DWORD size)
{
	CheckIndex(index);
	heap.Free(allItems[index]->features);
	allItems[index]->features = heap.Alloc((WORD)size);
	memcpy(allItems[index]->features, features, size);
	allItems[index]->featuresSize = size;
	if(opened)
	{
		UpdateRecord(index);
	}
}

void VFDDatabase::SetFingerID(int index, const CString &fingerId)
{
	CheckIndex(index);
	CopyFingerID(allItems[index]->fingerId, allItems[index]->fingerIdLength, fingerId, fingerId.GetLength());
	if(opened)
	{
		UpdateRecord(index);
	}
}

int VFDDatabase::Add(const BYTE * features, DWORD size, LPCTSTR fingerId)
{
	int index = AddItem(0, features, size, fingerId, sizeof(fingerId));
	if(opened)
	{
		SetRecord(index);
		try
		{
			TESTHR(recordBinding->AddNew(&record));
			allItems[index]->id = record.id;
		}
		catch(_com_error &e)
		{
			CString error;
			error.Format("Can not add record to the database : %s",e.ErrorMessage());
			AfxMessageBox(error);
			AfxThrowUserException();
		}
	}
	return index;
}

void VFDDatabase::Remove(int index)
{
	CheckIndex(index);
	VFDDBItem *item = allItems[index];
	int id = item->id;
	int g = VFFeatGetG(item->features); VFCheckResult(g);
	int i =  item->gIndex;
	int c = --count[g];
	memmove(items[g] + i, items[g] + i + 1, sizeof(VFDDBItem*) * (c - i));
	int k;
	for(k = i; k < c; k++)
		items[g][k]->gIndex--;
	int cp = VFDBCapacity(c);
	if(cp != capacity[g])
	{
		items[g] = (VFDDBItem**)realloc(items[g], sizeof(VFDDBItem*) * cp);
		capacity[g] = cp;
	}
	allCount--;
	memmove(allItems + index, allItems + index + 1, sizeof(VFDDBItem*) * (allCount - index));
	for(k = index; k < allCount; k++)
		allItems[k]->index--;
	cp = VFDBAllCapacity(allCount);
	if(allCapacity != cp)
	{
		allItems = (VFDDBItem**)realloc(allItems, sizeof(VFDDBItem*) * cp);
		allCapacity = cp;
	}
 	heap.Free(item->features);
	heap.Free((BYTE*)item);
	if(opened)
	{
		if(FindRecord(id))
			try
			{
				recordset->Delete(ADODB::adAffectCurrent);
			}
			catch(_com_error &e)
			{
				CString error;
				error.Format("Can not delete record from the database  : %s ",e.ErrorMessage());
				AfxMessageBox(error);
				AfxThrowUserException();
			}
	}
}

void VFDDatabase::FindStart(const BYTE * features, VFDBFindData &fd)
{
	fd.g = fd.testG = VFFeatGetG(features); VFCheckResult(fd.testG);
	fd.index = -1;
}

int VFDDatabase::FindNext(VFDBFindData &fd)
{
	fd.index++;
	while(fd.index >= count[fd.g])
	{
		fd.index = 0;
		int gDif = fd.g - fd.testG;
		fd.g = fd.testG + (gDif == 0 ? 1 : gDif > 0 ? -gDif : -gDif + 1);
		gDif = fd.g - fd.testG;
		if(fd.g >= 255)
		{
			fd.g = fd.testG - gDif;
			if(fd.g < 0) return -1;
		}
		else if(fd.g < 0)
		{
			fd.g = fd.testG - gDif + 1;
			if(fd.g >= 255) return -1;
		}
	}
	return items[fd.g][fd.index]->index;
}

void VFDDatabase::FindEnd(VFDBFindData &/*fd*/)
{
}

const CString defaultDatabaseProvider("Microsoft.Jet.OLEDB.4.0");
const CString defaultDatabaseFileExtension("mdb");

CString GetVFDBDefaultFilter()
{
	return "*." + defaultDatabaseFileExtension;
}

CString GetVFDBDefaultFilterName()
{
	return "Microsoft Access Databases";
}

CString MakeFileFilterString(const CString &name, const CString &filter, bool filter_in_name)
{
	CString rt;
	if(filter_in_name) {
	 rt.Format("%s (%s)",name,filter);
	} else {
	 rt.Format("%s|%s",name,filter);
	}
	return rt;	
}

CString GetVFDBDefaultFilter(bool filterInName)
{
	return MakeFileFilterString(GetVFDBDefaultFilterName(), GetVFDBDefaultFilter(), filterInName);
}
