#include "System.h"
#include "MainForm.h"
#include "Utils.h"
#include "Log.h"
#include "FileListForm.h"
#include "FileListEntryForm.h"
#include "AboutForm.h"
#include "OptionsForm.h"
#include "WaitForm.h"
#include "VFOptionsForm.h"

string VFDemoModeToInvariantString(VFDemoMode value)
{
	switch(value)
	{
		case vfdmNone: return TEXT("None");
		case vfdmEnrollment: return TEXT("Enrollment");
		case vfdmEnrollmentWithGen: return TEXT("EnrollmentWithGen");
		case vfdmIdentification: return TEXT("Identification");
		case vfdmVerification: return TEXT("Verification");
		default: throw exception("Can not convert VFDemoMode to string");
	}
}

VFDemoMode VFDemoModeFromInvariantString(const string &value)
{
	if(value == TEXT("None")) return vfdmNone;
	else if(value == TEXT("Enrollment")) return vfdmEnrollment;
	else if(value == TEXT("EnrollmentWithGen")) return vfdmEnrollmentWithGen;
	else if(value == TEXT("Identification")) return vfdmIdentification;
	else if(value == TEXT("Verification")) return vfdmVerification;
	else throw exception("Can not convert string to VFDemoMode");
}

string VFDemoModeToString(VFDemoMode value)
{
	switch(value)
	{
		case vfdmNone: return TEXT("None");
		case vfdmEnrollment: return TEXT("Enrollment");
		case vfdmEnrollmentWithGen: return TEXT("Enrollment with generalization");
		case vfdmIdentification: return TEXT("Identification");
		case vfdmVerification: return TEXT("Verification");
		default: return TEXT("[Unknown]");
	}
}

///////////////
//           //
// CMainForm //
//           //
///////////////

CMainForm::CMainForm(const string &appName, const string &engineName)
	: CForm(), image_from_scanner(false),
	task_thread(NULL), extraction_thread(NULL), identification_thread(NULL),
	stop_task(false), stop_identification(false),
	appName(appName), engineName(engineName),
	mode(vfdmNone), lastMode(vfdmEnrollment), changeBackMode(true),
	databaseFileName(ChangeFileExtension(appName, defaultDatabaseFileExtension)), createDatabase(true),
	settingsFileName(GetFullFileName(ChangeFileExtension(appName, TEXT("settings"))))
{
	database = new CVFDatabase();

	image = new CVFImage();
	image_after = new CVFImage();
	features = new CVFFeatures();

	int i;

	for(i = 0; i < VF_GENERALIZE_COUNT; i++)
	{
		gen_image[i] = new CVFImage();
		gen_image_after[i] = new CVFImage();
		gen_features[i] = new CVFFeatures();
	}
	gen_index = 0;

	for(i = 0; i < 2; i++)
	{
		ver_image[i] = new CVFImage();
		ver_image_after[i] = new CVFImage();
		ver_features[i] = new CVFFeatures();
	}
	ver_index = 0;

	file_list = new VFFileList();

	InitControls();

	open_dialog = new COpenFileDialog();
	open_dialog->AllowMultiSelect = true;
	open_dialog->Filter = CVFImageIO::GetLoadFileFilterString();
	open_list_dialog = new COpenFileDialog();
	open_list_dialog->Filter = GetLoadVFFileListFilterString();
	open_list_dialog->Title = TEXT("Open file list");
	save_dialog = new CSaveFileDialog();
	save_dialog->Filter = CVFImageIO::GetSaveFileFilterString();

	main_log = new CGUILog();
	main_log->SetEditControl(richedit_main);

	scanner_log = new CGUILog();
	scanner_log->SetEditControl(richedit_scanner);

	AddVersionToLog(main_log, Application()->Name(), Application()->VersionHigh(), Application()->VersionLow());
	main_log->WriteLine(TEXT("OS: %s. Version %lu.%lu (Build %lu). Service Pack Version %lu.%lu"),
		(LPCTSTR)Application()->OSName(),
		Application()->OSMajorVersion(), Application()->OSMinorVersion(), Application()->OSBuild(),
		Application()->OSSPMajorVersion(), Application()->OSSPMinorVersion());
	AddVersionToLog(main_log, VFGetName(), VFGetVersionHigh(), VFGetVersionLow());
	main_log->NewLine();

	AddVersionToLog(scanner_log, SMGetName(), SMGetVersionHigh(), SMGetVersionLow());
	scanner_log->NewLine();

	UpdateTaskProgress(false, 0);
	UpdateIdentificationProgress(false, 0);

	UpdateStatus();
}

void CMainForm::InitControls()
{
	SetFormState(fsMaximized);
	SetFormPosition(fpDefault);

	view_left = new CVFView();
	view_right = new CVFView();
	view_right->MinColor = RGB(0, 230, 0);
	
	captionpanel_main = new CCaptionPanel();
	captionpanel_main->SetText(STRING("Main log"));
	panel_main = new CPanel();
	label_main = new CLabel();
	label_main->SetLocation(8, 8);
	progressbar_main = new CProgressBar();
	progressbar_main->SetLocation(8, label_main->Bottom() + 8);
	label_main_time = new CLabel();
	label_main_time->SetLocation(8, progressbar_main->Bottom() + 8);
	panel_main->SetClientHeight(label_main_time->Bottom() + 8);
	button_main_stop = new CButton();
	button_main_stop->SetVCenter(panel_main->ClientHeight() / 2);
	button_main_stop->SetText(STRING("Stop"));
	panel_main->AddControl(label_main);
	panel_main->AddControl(progressbar_main);
	panel_main->AddControl(label_main_time);
	panel_main->AddControl(button_main_stop);
	richedit_main = new CRichEdit();
	richedit_main->SetReadOnly(true);
	captionpanel_main->AddControl(panel_main);
	richedit_main->SetTop(panel_main->Bottom());
	captionpanel_main->AddControl(richedit_main);

	captionpanel_scanner = new CCaptionPanel();
	captionpanel_scanner->SetText(STRING("Scanner log"));
	richedit_scanner = new CRichEdit();
	richedit_scanner->SetReadOnly(true);
	captionpanel_scanner->AddControl(richedit_scanner);

	captionpanel_results = new CCaptionPanel();
	captionpanel_results->SetText(STRING("Identification results"));
	panel_results = new CPanel();
	progressbar_results = new CProgressBar();
	progressbar_results->SetLocation(8, 8);
	label_results_time = new CLabel();
	label_results_time->SetLocation(8, progressbar_results->Bottom() + 8);
	button_results_stop = new CButton();
	button_results_stop->SetText(STRING("Stop"));
	panel_results->SetClientHeight(label_results_time->Bottom() + 8);
	button_results_stop->SetVCenter(panel_results->ClientHeight() / 2);
	panel_results->AddControl(progressbar_results);
	panel_results->AddControl(label_results_time);
	panel_results->AddControl(button_results_stop);
	listview_results = new CListView();
	listview_results->SetNoSortHeader(true);
	listview_results->SetSingleSel(true);
	listview_results->SetFullRowSelect(true);
	listview_results->AddColumn(STRING("G"), 30);
	listview_results->AddColumn(STRING("Result"), 50);
	listview_results->AddColumn(STRING("Similarity"), 60);
	listview_results->AddColumn(STRING("Record ID"), 70);
	listview_results->AddColumn(STRING("Finger ID"), 250);
	captionpanel_results->AddControl(panel_results);
	listview_results->SetTop(panel_results->Bottom());
	captionpanel_results->AddControl(listview_results);

	statusbar = new CStatusBar();
	statusbar->AddPart(250);
	statusbar->AddPart(-1);

	splitter_views = new CVSplitter();
	splitter_horz = new CHSplitter();
	splitter_horz->SetNextControl(statusbar);
	splitter_horz->SetRatio(0.6f);
	splitter_v1 = new CVSplitter();
	splitter_v1->SetRatio(0.4f);
	splitter_v2 = new CVSplitter();
	splitter_v2->SetPrevControl(splitter_v1);
	splitter_v2->SetRatio(1.0f / 3.0f);

	AddControl(view_left);
	AddControl(splitter_views);
	AddControl(view_right);
	AddControl(splitter_horz);
	AddControl(captionpanel_main);
	AddControl(splitter_v1);
	AddControl(captionpanel_scanner);
	AddControl(splitter_v2);
	AddControl(captionpanel_results);
	AddControl(statusbar);
}

CMainForm::~CMainForm()
{
	delete scanner_log;
	delete main_log;

	delete save_dialog;
	delete open_list_dialog;
	delete open_dialog;

	delete file_list;

	int i;

	for(i = 0; i < VF_GENERALIZE_COUNT; i++)
	{
		delete gen_image[i];
		delete gen_image_after[i];
		delete gen_features[i];
	}

	for(i = 0; i < 2; i++)
	{
		delete ver_image[i];
		delete ver_image_after[i];
		delete ver_features[i];
	}

	delete image;
	delete image_after;
	delete features;

	delete database;
}

void CMainForm::LoadSettings()
{
	try
	{
		VFInitDefaultOptions();
		vfOptions = vfDefaultOptions;
		VFDemoInitDefaultOptions();
		options = vfDemoDefaultOptions;
		MSXML2::IXMLDOMDocumentPtr document = XmlCreateDocument();
		if(FileExists(settingsFileName))
		{
			if(!document->load((_bstr_t)settingsFileName)) throw exception("Can not load settings from file");
			LoadSettings(document);
		}
	}
	catch(exception &e)
	{
		ErrorBox(this, e, TEXT("Loading settings"));
	}
	catch(_com_error &e)
	{
		ErrorBox(this, e, TEXT("Loading settings"));
	}
}

void CMainForm::SaveSettings()
{
	try
	{
		MSXML2::IXMLDOMDocumentPtr document = XmlCreateDocument();
		SaveSettings(document);
		document->preserveWhiteSpace = true;
		document->save((_bstr_t)settingsFileName);
	}
	catch(exception &e)
	{
		ErrorBox(this, e, TEXT("Saving settings"));
	}
	catch(_com_error &e)
	{
		ErrorBox(this, e, TEXT("Saving settings"));
	}
}

void CMainForm::LoadSettings(MSXML2::IXMLDOMDocumentPtr document)
{
	MSXML2::IXMLDOMElementPtr node = document->documentElement;
	if(!(bool)node || node->nodeName != appName) throw EXCEPTION(Format("<%s> element not found", (LPCTSTR)appName));
	
	MSXML2::IXMLDOMNodePtr veriFingerNode = node->selectSingleNode(engineName);
	if((bool)veriFingerNode)
	{
		try
		{
			MSXML2::IXMLDOMNodePtr vfOptionsNode = veriFingerNode->selectSingleNode("Options");
			if((bool)vfOptionsNode)
			{
				VFLoadOptions(vfOptionsNode, vfOptions);
				VFUpdateOptions(vfOptions);
			}
		}
		catch(exception &e)
		{
			ErrorBox(this, e, TEXT("Loading ") + engineName + TEXT(" options"));
		}
		catch(_com_error &e)
		{
			ErrorBox(this, e, TEXT("Loading ") + engineName + TEXT(" options"));
		}
	}
	
	try
	{
		MSXML2::IXMLDOMNodePtr optionsNode = node->selectSingleNode("Options");
		if((bool)optionsNode)
			VFDemoLoadOptions(optionsNode, options);
	}
	catch(exception &e)
	{
		ErrorBox(this, e, TEXT("Loading options"));
	}
	catch(_com_error &e)
	{
		ErrorBox(this, e, TEXT("Loading options"));
	}
	
	try
	{
		MSXML2::IXMLDOMNodePtr mainFormNode = node->selectSingleNode("MainForm");
		if((bool)mainFormNode)
			LoadOptions(mainFormNode);
	}
	catch(exception &e)
	{
		ErrorBox(this, e, TEXT("Loading Main form options"));
	}
	catch(_com_error &e)
	{
		ErrorBox(this, e, TEXT("Loading Main form options"));
	}
}

void CMainForm::SaveSettings(MSXML2::IXMLDOMDocumentPtr document)
{
	MSXML2::IXMLDOMElementPtr node = document->createElement(appName);
	
	MSXML2::IXMLDOMElementPtr veriFingerNode = document->createElement(engineName);
	MSXML2::IXMLDOMElementPtr vfOptionsNode = document->createElement("Options");
	VFSaveOptions(document, vfOptionsNode, vfOptions);
	veriFingerNode->appendChild(vfOptionsNode);
	node->appendChild(veriFingerNode);
	
	MSXML2::IXMLDOMElementPtr optionsNode = document->createElement("Options");
	VFDemoSaveOptions(document, optionsNode, options);
	node->appendChild(optionsNode);
	
	MSXML2::IXMLDOMElementPtr mainFormNode = document->createElement("MainForm");
	SaveOptions(document, mainFormNode);
	node->appendChild(mainFormNode);
	
	document->documentElement = node;
}

void CMainForm::LoadOptions(MSXML2::IXMLDOMNodePtr node)
{
	MSXML2::IXMLDOMNamedNodeMapPtr attributes = node->attributes;
	XmlGetAttributeValue(attributes, FilesDirectory, string);
	XmlGetAttributeValue(attributes, FileListsDirectory, string);
	XmlGetAttributeValueN(attributes, lastMode, "Mode", VFDemoModeFromInvariantString);

	MSXML2::IXMLDOMNodePtr databaseNode = node->selectSingleNode("Database");
	if((bool)databaseNode)
	{
		MSXML2::IXMLDOMNamedNodeMapPtr databaseAttributes = databaseNode->attributes;
		XmlGetAttributeValueN(databaseAttributes, databaseFileName, "FileName", string);
		XmlGetAttributeValueN(databaseAttributes, createDatabase, "CreateIfNotExists", StringToBool);
	}

	MSXML2::IXMLDOMElementPtr viewNode = node->selectSingleNode("View");
	if((bool)viewNode)
	{
		MSXML2::IXMLDOMNamedNodeMapPtr viewAttributes = viewNode->attributes;
		XmlGetAttributeMemberPValue(viewAttributes, view_right, ShowFeatures, StringToBool);
		XmlGetAttributeMemberPValue(viewAttributes, view_right, ShowSingularPoints, StringToBool);
		XmlGetAttributeMemberPValue(viewAttributes, view_right, ShowBlockedOrientations, StringToBool);
		UpdateViewMenu();
	}
}

void CMainForm::SaveOptions(MSXML2::IXMLDOMDocumentPtr document, MSXML2::IXMLDOMNodePtr node)
{
	MSXML2::IXMLDOMNamedNodeMapPtr attributes = node->attributes;
	XmlSetAttributeValue(document, attributes, FilesDirectory, string);
	XmlSetAttributeValue(document, attributes, FileListsDirectory, string);
	XmlSetAttributeValue(document, attributes, Mode, VFDemoModeToInvariantString);

	MSXML2::IXMLDOMElementPtr databaseNode = document->createElement("Database");
	{
		MSXML2::IXMLDOMNamedNodeMapPtr databaseAttributes = databaseNode->attributes;
		XmlSetAttributeValueN(document, databaseAttributes, DatabaseFileName, "FileName", string);
		XmlSetAttributeValueN(document, databaseAttributes, CreateDatabase, "CreateIfNotExists", BoolToString);
	}
	node->appendChild(databaseNode);

	MSXML2::IXMLDOMElementPtr viewNode = document->createElement("View");
	{
		MSXML2::IXMLDOMNamedNodeMapPtr viewAttributes = viewNode->attributes;
		XmlSetAttributeMemberPValue(document, viewAttributes, view_right, ShowFeatures, BoolToString);
		XmlSetAttributeMemberPValue(document, viewAttributes, view_right, ShowSingularPoints, BoolToString);
		XmlSetAttributeMemberPValue(document, viewAttributes, view_right, ShowBlockedOrientations, BoolToString);
	}
	node->appendChild(viewNode);
}

void CMainForm::OnCreate()
{
	CForm::OnCreate();

	LoadSettings();

	SetMenuItemText(ID_TOOLS_ENGINEOPTIONS, TEXT('&') + engineName + TEXT(" options..."));

	UpdateMode();
	UpdateViewMenu();

	INT result = SMSetMonitor(SMMonitorProc, this);
	scanner_log->WriteLine(TEXT("SetMonitor: %s"), (LPCTSTR)SMResultToString(result));
	scanner_log->NewLine();
	strings ids;
	SMGetScannerIds(ids);
	scanner_log->WriteLine(IntToString(ids.Length(), STRING("scanner"), STRING("scanners")) + TEXT(" found"));
	scanner_log->NewLine();
	for(string *i = ids.Begin(); i < ids.End(); i++)
		scanner_log->WriteLine(TEXT("Starting capturing from scanner %s: %s"), (LPCTSTR)*i,
			(LPCTSTR)SMResultToString(SMStartCapturing(*i, SMImageProc, SMStateProc, this)));

	if(!scanner_log->LastLine().IsEmpty()) scanner_log->NewLine();
}

void CMainForm::OnDestroy()
{
	SMStopCapturing();
	SMRemoveMonitor();

	SaveSettings();

	CForm::OnDestroy();
}

void CMainForm::OnCommand(WORD id, WORD event, CControl *control)
{
	if(control)
	{
		if(event == SN_SPLITTED)
		{
			if(control == splitter_views) AdjustControls(true, false);
			else if(control == splitter_horz) AdjustControls(true, true);
			else if(control == splitter_v1 || control == splitter_v2) AdjustControls(false, true);
		}
		if(event == BN_CLICKED)
		{
			if(control == button_main_stop) StopTask();
			else if(control == button_results_stop) StopIdentification();
		}
	}
	else
		switch(id)
		{
			case ID_FILE_CLEAR: Clear(); break;
			case ID_FILE_OPEN: OnFileOpen(); break;
			case ID_FILE_OPENFILELIST: OnFileOpenFileList(); break;
			case ID_FILE_SAVE: OnFileSave(); break;
			case ID_FILE_SAVERIGHT: OnFileSaveRight(); break;
			case ID_FILE_STOP: OnFileStop(); break;
			case ID_FILE_EXIT: OnFileExit(); break;
			case ID_MODE_ENROLLMENT: OnModeEnrollment(); break;
			case ID_MODE_ENROLLMENTWITHGEN: OnModeEnrollmentWithGen(); break;
			case ID_MODE_IDENTIFICATION: OnModeIdentification(); break;
			case ID_MODE_VERIFICATION: OnModeVerification(); break;
			case ID_VIEW_ZOOMIN: OnViewZoomIn(); break;
			case ID_VIEW_ZOOMOUT: OnViewZoomOut(); break;
			case ID_VIEW_ZOOM1TO1: OnViewZoom1To1(); break;
			case ID_VIEW_SHOWFEATURES: OnViewShowFeatures(); break;
			case ID_VIEW_SHOWSINGULARPOINTS: OnViewShowSP(); break;
			case ID_VIEW_SHOWBLOCKEDORIENTATIONS: OnViewShowBO(); break;
			case ID_TOOLS_FILELISTEDITOR: OnToolsFileListEditor(); break;
			case ID_TOOLS_CLEARDATABASE: OnToolsClearDatabase(); break;
			case ID_TOOLS_CLEARMAINLOG: OnToolsClearMainLog(); break;
			case ID_TOOLS_CLEARSCANNERLOG: OnToolsClearScannerLog(); break;
			case ID_TOOLS_OPTIONS: OnToolsOptions(); break;
			case ID_TOOLS_ENGINEOPTIONS: OnToolsEngineOptions(); break;
			case ID_HELP_ABOUT: OnHelpAbout(); break;
		}
}

LRESULT CMainForm::OnMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case UM_TASK_STATE: OnTaskState(string((LPCTSTR)lParam)); break;
		case UM_TASK_PROGRESS: return OnTaskProgress(wParam != 0, lParam);
		case UM_IDENTIFICATION_PROGRESS: return OnIdentificationProgress(wParam != 0, (IdenProgress*)lParam);
		case CM_THREAD_FINISHED:
			{
				CThread *thread = (CThread*)lParam;
				if(thread == task_thread) OnTaskComplete();
				else if(thread == extraction_thread) OnExtractionComplete();
				else if(thread == identification_thread) OnIdentificationComplete();
			}
			break;
		default: return CForm::OnMessage(message, wParam, lParam);
	}
	return 0;
}

void CMainForm::OnTaskState(const string &state)
{
	UpdateTaskProgress(false, 0, state);
}

bool CMainForm::OnTaskProgress(bool started, int value)
{
	if(stop_task) return false;
	UpdateTaskProgress(started, value);
	if(DBOpenRunning() && started) UpdateStatus();
	else if(FileListRunning() && started) OnFileListImage(value);
	return true;
}

void CMainForm::OnDBOpenComplete(DWORD time)
{
	CDBOpenThread *db_open_thread = dynamic_cast<CDBOpenThread*>(task_thread);
	if(db_open_thread->Failed())
	{
		main_log->WriteLine(TEXT(". Database not opened"));
		ErrorBox(this, db_open_thread->Error());
	}
	else
	{
		int c = database->Count;
		main_log->WriteLine(TEXT(": %s loaded"), (LPCTSTR)IntToString(c, STRING("record"), STRING("records"), STRING("no")));
		if(c)
		{
			main_log->WriteLine(TEXT("Time: %s%s"), (LPCTSTR)TimeToString(time),
				time ? (LPCTSTR)Format(TEXT(" (%u records per second)"), c * 1000 / time) : TEXT(""));
		}
	}
	UpdateStatus();
}

void CMainForm::CloseDB()
{
	if(database->Opened())
	{
		CDBCloseThread *thread = new CDBCloseThread(NULL, database);
		DWORD tc = GetTickCount();
		WaitBox(this, thread, thread->TaskString());
		tc = GetTickCount() - tc;
		if(thread->Failed())
		{
			main_log->WriteLine(thread->TaskName() + TEXT(" failed"));
			ErrorBox(this, thread->Error());
		}
		else main_log->WriteLine(TEXT("Database closed. Time: %s"), (LPCTSTR)TimeToString(tc));
		main_log->NewLine();
		UpdateStatus();
	}
}

void CMainForm::ReOpenDB()
{
	CloseDB();
	StartDBOpen();
}

void CMainForm::OnFileOpen()
{
	if(!TaskRunning() && mode != vfdmNone)
	{
		open_dialog->InitialDirectory = filesDirectory;
		open_dialog->FileName.Clear();
		if(open_dialog->Show(this) == IDOK)
		{
			filesDirectory = GetDirectoryName(open_dialog->FileName);
			if(mode == vfdmEnrollmentWithGen)
				if(open_dialog->FileNames.Length() != VF_GENERALIZE_COUNT)
				{
					MessageBox(this, Format(TEXT("%s requires %d images"), (LPCTSTR)GetModeString(), VF_GENERALIZE_COUNT), MB_OK | MB_ICONERROR);
					return;
				}
				else gen_index = 0;
			else if(mode == vfdmVerification)
				if(open_dialog->FileNames.Length() != 2)
				{
					MessageBox(this, Format(TEXT("%s verification requires 2 images"), (LPCTSTR)GetModeString()), MB_OK | MB_ICONERROR);
					return;
				}
				else ver_index = 0;
			bool ok;
			if(options.imagesFromFile.prompt && (mode == vfdmEnrollment || mode == vfdmEnrollmentWithGen || mode == vfdmIdentification))
			{
				DWORD flags = mode == vfdmIdentification ? FLEB_IDENTIFY : FLEB_ENROLL;
				if(open_dialog->FileNames.Length() > 1) SetFlag(flags, FLEB_MANY_ENTRIES);
				string filename = open_dialog->FileName;
				ok = FileListEntryBox(this, flags, filename, options.imagesFromFile.resolution, options.imagesFromFile.fingerId);
			}
			else ok = true;
			if(ok)
			{
				file_list->Clear();
				for(string *i = open_dialog->FileNames.Begin(); i < open_dialog->FileNames.End(); i++)
					file_list->Add(VFFileListItem(*i, options.imagesFromFile.resolution, MakeFingerID(options.imagesFromFile.fingerId, *i, false)));
				StartFileList();
			}
		}
	}
}

void CMainForm::OnFileOpenFileList()
{
	if(!TaskRunning() && (mode == vfdmEnrollment || mode == vfdmIdentification))
	{
		open_list_dialog->InitialDirectory = fileListsDirectory;
		open_list_dialog->FileName.Clear();
		if(open_list_dialog->Show(this) == IDOK)
		{
			fileListsDirectory = GetDirectoryName(open_list_dialog->FileName);
			if(LoadVFFileList(open_list_dialog->FileName, file_list))
			{
				if(file_list->IsEmpty()) MessageBox(this, TEXT("File list is empty"), MB_OK | MB_ICONERROR);
				else StartFileList();
			}
			else MessageBox(this, Format(TEXT("Can not open file list \"%s\""), (LPCTSTR)open_list_dialog->FileName), MB_OK | MB_ICONERROR);
		}
	}
}

void CMainForm::OnFileSave()
{
	if(!TaskRunning())
	{
		SaveImage(image);
	}
}

void CMainForm::OnFileSaveRight()
{
	if(!TaskRunning())
	{
		SaveImage(image_after);
	}
}

void CMainForm::SaveImage(CVFImage *image)
{
	if(image->IsEmpty()) return;
	save_dialog->FileName.Clear();
	if(save_dialog->Show(this) == IDOK)
	{
		if(!image->SaveToFile(save_dialog->FileName, save_dialog->FilterIndex))
			MessageBox(this, Format(TEXT("Can not save image to file \"%s\""), save_dialog->FileName), MB_OK | MB_ICONERROR);
	}
}

void CMainForm::OnFileStop()
{
	StopTask();
}

void CMainForm::OnFileExit()
{
	Close();
}

void CMainForm::StartFileList()
{
	if(file_list->Length() == 1)
	{
		VFFileListItem *item = file_list[0];
		OnFileImage(item->filename, item->resolution, item->finger_id);
		return;
	}
	main_log->Write(TEXT("%d files selected"), file_list->Length());
	count = 0;
	count_rej = 0;
	total_size = 0;
	total_ext_time = 0;
	total_iden = 0;
	total_iden_time = 0;
	fingers_found = 0;
	false_accepted = 0;
	false_rejected = 0;
	StartTask(new CFileListThread(this, file_list));
}

void CMainForm::OnFileListImage(int file_index)
{
	task_thread->Suspend();
	if(file_index == 0) main_log->NewLine();
	main_log->WriteLine(TEXT("File %d of %d"), file_index + 1, file_list->Length());
	VFFileListItem &item = (*file_list)[file_index];
	if(OnFileImage(item.filename, item.resolution, item.finger_id))
	{
		// if this is the first file in the file list then show input image and clear output view
		if(file_index == 0) 
		{
			view_left->Image = image;
			view_left->Resolution = resolution;
			view_right->Image = NULL;
			view_right->Features = NULL;
			listview_results->ClearItems();
		}
	}
	else task_thread->Resume();
}

void CMainForm::OnFileListComplete(DWORD time)
{
	main_log->WriteLine(TEXT(": %s proceeded"), (LPCTSTR)IntToString(count, STRING("file"), STRING("files"), STRING("no")));
	if(count)
	{
		if(mode == vfdmEnrollment || mode == vfdmIdentification)
		{
			main_log->WriteLine(TEXT("Average time per file: %s (%s total)"), (LPCTSTR)TimeToString(time / count),
				(LPCTSTR)TimeToString(time));
			main_log->WriteLine(TEXT("Average extraction time: %s (%s total)"), (LPCTSTR)TimeToString(total_ext_time / count),
				(LPCTSTR)TimeToString(total_ext_time));
			int count_left = count - count_rej;
			main_log->Write(TEXT("Images rejected: %d (%d%%)"), count_rej, Percent(count_rej, count));
			if(count_left)
			{
				main_log->WriteLine(TEXT(". For %d non-rejected:"), count_left);
				if(mode == vfdmEnrollment)
				{
					main_log->WriteLine(TEXT("Average features size: %s (%s total)"),
						(LPCTSTR)SizeToString(total_size / count_left),
						(LPCTSTR)SizeToString(total_size));
				}
				else
				{
					main_log->WriteLine(TEXT("Average matching time: %s (%s total)"),
						(LPCTSTR)TimeToString(total_iden_time / count_left),
						(LPCTSTR)TimeToString(total_iden_time));
					main_log->WriteLine(TEXT("Average matching speed: %s"),
						(LPCTSTR)MatchingSpeedToString(total_iden, total_iden_time));
					if(options.identification.useG)
					{
						main_log->WriteLine(TEXT("Average effective matching speed: %s"),
							(LPCTSTR)MatchingSpeedToString(count_left * database->Count, total_iden_time));
						main_log->Write(TEXT("Average fingerprints matched: %d"), total_iden / count_left);
						if(database->Count)
						{
							double p = PercentF(total_iden, count_left * database->Count);
							if(options.identification.matchAllDatabase && p == 100.0) main_log->WriteLine(TEXT(" (all database)"));
							else main_log->WriteLine(TEXT(" (%d%% of database size)"), Round(p));
						}
						else main_log->NewLine();
					}
					main_log->WriteLine(TEXT("FAR: %s (%d of %d)"), (LPCTSTR)PercentToString(false_accepted, count_left - fingers_found),
						false_accepted, count_left - fingers_found);
					main_log->WriteLine(TEXT("FRR: %s (%d of %d)"), (LPCTSTR)PercentToString(false_rejected, fingers_found),
						false_rejected, fingers_found);
				}
			}
		}
		else main_log->WriteLine(TEXT("Time: %s"), (LPCTSTR)TimeToString(time));
	}
}

bool CMainForm::OnFileImage(const string &filename, int resolution, const string &finger_id)
{
	this->resolution = resolution;
	main_log->WriteLine(TEXT("Loading image: %s"), (LPCTSTR)filename);
	if(!image->LoadFromFile(filename))
	{
		view_left->Image = NULL;
		view_right->Image = NULL;
		view_right->Features = NULL;
		main_log->WriteLine(TEXT("Can not load image, further processing not performed"));
		main_log->NewLine();
		ErrorPrompt(Format(TEXT("Can not load image from file \"%s\""), (LPCTSTR)filename));
		return false;
	}
	else
	{
		this->finger_id = finger_id;
		OnImage();
		return true;
	}
}

void CMainForm::OnSMMonitor(const string & scanner_id, DWORD event)
{
	if(!scanner_log->LastLine().IsEmpty()) scanner_log->NewLine();
	scanner_log->WriteLine(SMEventToString(event, scanner_id));
	scanner_log->NewLine();
}

void CMainForm::OnScannerImage(const string & scanner_id, int width, int height, const BYTE * image, int resolution)
{
	if(!TaskRunning() && mode != vfdmNone)
	{
		this->scanner_id = scanner_id;
		this->resolution = resolution;
		finger_id.Clear();
		this->image->Assign(width, height, image);
		image_from_scanner = true;
		main_log->WriteLine(TEXT("Image from scanner: %s"), (LPCTSTR)this->scanner_id);
		OnImage();
	}
}

void CMainForm::OnScannerState(const string & scanner_id, DWORD state)
{
	scanner_log->WriteLine(SMStateToString(state, scanner_id));
}

void CMainForm::OnImage()
{
	// if not processing file list then show the input image and clear output view
	if(!FileListRunning())
	{
		view_left->Image = image;
		view_left->Resolution = resolution;
		view_right->Image = NULL;
		view_right->Features = NULL;
		listview_results->ClearItems();
	}
	main_log->Write(TEXT("Size: %dx%d. Resolution: %d dpi"), image->Width(), image->Height(), resolution);
	main_log->WriteLine(TEXT(". Extraction started..."));
	extraction_thread = new CExtractionThread(this, resolution, image, image_after);
	UpdateTask();
	extraction_thread->Resume();
}

void CMainForm::OnExtractionComplete()
{
	INT result = extraction_thread->Result();
	bool ok;
	if(VFSucceeded(result) || result == VFE_LOW_QUALITY_IMAGE)
	{
		main_log->WriteLine(TEXT("Extraction complete: %s. Time: %s"), (LPCTSTR)VFResultToString(result),
			(LPCTSTR)TimeToString(extraction_thread->Time()));
		ok = result != VFE_LOW_QUALITY_IMAGE || options.useLowQualityImages;
		features->SetData(extraction_thread->Features(), extraction_thread->FeaturesSize());
		// show output
		view_right->Image = image_after;
		view_right->Resolution = resolution;
		view_right->Features = features;
	}
	else
	{
		view_right->Image = NULL;
		view_right->Features = NULL;
		main_log->WriteLine(TEXT("Extraction error: %s. Futher processing not performed"), (LPCTSTR)VFErrorToString(result));
		ok = false;
		ErrorPrompt(Format(TEXT("Extraction error: %s"), (LPCTSTR)VFErrorToString(result)));
	}
	if(FileListRunning())
	{
		count++;
		total_ext_time += extraction_thread->Time();
		if(ok)
		{
			total_size += extraction_thread->FeaturesSize();
		}
		else count_rej++;
	}
	// if processing file list then input image is not shown yet and results not cleared
	if(FileListRunning())
	{
		view_left->Image = image;
		view_left->Resolution = resolution;
		listview_results->ClearItems();
	}
	if(mode == vfdmEnrollmentWithGen)
	{
		if(!ok)
		{
			main_log->NewLine();
			main_log->WriteLine(TEXT("Generalization aborted"));
			gen_index = 0;
			if(FileListRunning()) stop_task = true;
		}
		else
		{
			gen_resolution[gen_index] = resolution;
			gen_image[gen_index]->Assign(image);
			gen_image_after[gen_index]->Assign(image_after);
			gen_features[gen_index]->Assign(features);
			gen_finger_id[gen_index] = finger_id;
			gen_index++;
			if(gen_index == VF_GENERALIZE_COUNT)
			{
				gen_index = 0;
			}
			else ok = false;
		}
	}
	else if(mode == vfdmVerification)
	{
		if(!ok)
		{
			main_log->NewLine();
			main_log->WriteLine(TEXT("Verification aborted"));
			ver_index = 0;
			if(FileListRunning()) stop_task = true;
		}
		else
		{
			ver_resolution[ver_index] = resolution;
			ver_image[ver_index]->Assign(image);
			ver_image_after[ver_index]->Assign(image);
			ver_features[ver_index]->Assign(features);
			ver_finger_id[ver_index] = finger_id;
			ver_index++;
			if(ver_index == 2)
			{
				ver_index = 0;
			}
			else ok = false;
		}
	}
	bool resume = true;
	if(ok)
		switch(mode)
		{
			case vfdmEnrollment:
				Enroll();
				break;
			case vfdmEnrollmentWithGen:
				EnrollWithGen();
				break;
			case vfdmIdentification:
				resume = !StartIdentification();
				break;
			case vfdmVerification:
				Verify();
				break;
		}
	image_from_scanner = false;
	delete extraction_thread;
	extraction_thread = NULL;
	if(resume)
	{
		UpdateTask();
		main_log->NewLine();
		if(FileListRunning()) task_thread->Resume();
	}
}

void CMainForm::Enroll()
{
	if(options.enrollment.checkMinimalMinutiaCount && features->Minutiae.Length() < options.enrollment.minimalMinutiaCount)
	{
		main_log->WriteLine(TEXT("Minutia count (%d) is less than minimal accepted value (%d)"), features->Minutiae.Length(), options.enrollment.minimalMinutiaCount);
		ErrorPrompt(Format(TEXT("Minutia count (%d) is less than minimal accepted value (%d)"),
			features->Minutiae.Length(), options.enrollment.minimalMinutiaCount));
		return;
	}
	if(image_from_scanner)
	{
		if(options.imagesFromScanner.prompt)
		{
			DWORD flags = FLEB_ENROLL | FLEB_FROM_SCANNER;
			if(!FileListEntryBox(this, flags, scanner_id, resolution, options.imagesFromScanner.fingerId)) return;
		}
		finger_id = MakeFingerID(options.imagesFromScanner.fingerId, scanner_id, true);
	}
	BYTE feat[VF_MAX_FEATURES_SIZE];
	DWORD size = features->GetData(feat);
	if(options.enrollment.searchForDuplicates)
	{
		CIdentificationThread * thread = new CIdentificationThread(this, feat, size, database, true, 100, false);
		WaitBox(this, thread, STRING("Searching for duplicates"));
		int index = thread->Index();
		delete thread;
		if(index != -1)
		{
			string dup_finger_id = database->FingerID(index);
			if(!options.enrollment.warnDifferentFingerId || finger_id != dup_finger_id)
			{
				main_log->WriteLine(TEXT("Duplicate found. Finger ID: \"%s\". Record ID: %d"), (LPCTSTR)dup_finger_id, database->ID(index));
				if(!ErrorPrompt(Format(TEXT("Duplicate found. Finger ID: \"%s\". Record ID: %d\nAdd to the database?"),
					(LPCTSTR)dup_finger_id, database->ID(index)), true)) return;
			}
		}
	}
	main_log->WriteLine(TEXT("Adding to the database. Features size: %s. Finger ID: \"%s\""),
		(LPCTSTR)SizeToString(size), (LPCTSTR)finger_id);
	try
	{
		database->Add(feat, size, finger_id);
		UpdateStatus();
	}
	catch(exception &e)
	{
		main_log->WriteLine(TEXT("Can not add to the database"));
		ErrorPrompt(Format(TEXT("Can not add to the database\n%s"), (LPCTSTR)AnsiToString(e.what())));
	}
}

void CMainForm::EnrollWithGen()
{
	main_log->NewLine();
	BYTE * gen_feat[VF_GENERALIZE_COUNT];
	int i;
	for(i = 0; i < VF_GENERALIZE_COUNT; i++)
	{
		gen_feat[i] = new BYTE[VF_MAX_FEATURES_SIZE];
		gen_features[i]->GetData(gen_feat[i]);
	}
	BYTE feat[VF_MAX_FEATURES_SIZE];
	DWORD size;
	double tc = GetTickCountEx();
	INT result = VFGeneralize(VF_GENERALIZE_COUNT, gen_feat, feat, &size, NULL);
	tc = GetTickCountEx() - tc;
	for(i = 0; i < VF_GENERALIZE_COUNT; i++)
		delete[] gen_feat[i];
	if(VFSucceeded(result) || result == VFE_FAILED)
	{
		main_log->WriteLine(TEXT("Generalization %s. Time: %s"), result == VFE_FAILED ? TEXT("failed")
			: (LPCTSTR)Format(TEXT("succeeded (based on image %d)"), result + 1), (LPCTSTR)TimeToString(tc));
		if(result != VFE_FAILED)
		{
			resolution = gen_resolution[result];
			image->Assign(gen_image[result]);
			image_after->Assign(gen_image_after[result]);
			features->SetData(feat, size);
			finger_id = gen_finger_id[result];
			view_left->Resolution = resolution;
			view_left->UpdateImage();
			view_right->Resolution = resolution;
			view_right->UpdateImage();
			view_right->UpdateFeatures();
			Enroll();
		}
	}
	else
	{
		string s = Format(TEXT("Generalization error: %s"), (LPCTSTR)VFErrorToString(result));
		main_log->WriteLine(s);
		ErrorPrompt(s);
	}
}

void CMainForm::Verify()
{
	main_log->NewLine();
	BYTE feat1[VF_MAX_FEATURES_SIZE], feat2[VF_MAX_FEATURES_SIZE];
	VFCheckResult(ver_features[0]->GetData(feat1));
	VFCheckResult(ver_features[1]->GetData(feat2));
	VFMatchDetails md;
	md.Size = sizeof(md);
	double tc = GetTickCountEx();
	INT result = VFVerify(feat1, feat2, &md, NULL);
	tc = GetTickCountEx() - tc;
	if(VFSucceeded(result) || result == VFE_FAILED)
	{
		main_log->WriteLine(TEXT("Verification %s. Similarity: %d. Time: %s"),
			result == VFE_FAILED ? TEXT("failed") : TEXT("succeeded"), md.Similarity,
			(LPCTSTR)TimeToString(tc));
	}
	else
	{
		string s = Format(TEXT("Verification error: %s"), (LPCTSTR)VFErrorToString(result));
		main_log->WriteLine(s);
		ErrorPrompt(s);
	}
}

bool CMainForm::StartIdentification()
{
	if(options.identification.checkForEmptyDatabase && !database->Count)
	{
		main_log->WriteLine(TEXT("Database is empty. Do not starting identification"));
		ErrorPrompt(STRING("Database is empty"));
		return false;
	}
	if(image_from_scanner)
	{
		if(options.imagesFromScanner.prompt)
		{
			DWORD flags = FLEB_ENROLL | FLEB_FROM_SCANNER;
			if(!FileListEntryBox(this, flags, scanner_id, resolution, options.imagesFromScanner.fingerId)) return false;
		}
		finger_id = MakeFingerID(options.imagesFromScanner.fingerId, scanner_id, true);
	}
	main_log->Write(TEXT("Finger ID: \"%s\""), (LPCTSTR)finger_id);
	if(options.identification.useG)
	{
		main_log->Write(TEXT(". G: %d"), features->G);
	}
	main_log->WriteLine(TEXT(". Identification started..."));
	count_iden = 0;
	finger_found = false;
	BYTE feat[VF_MAX_FEATURES_SIZE];
	DWORD size = features->GetData(feat);
	identification_thread = new CIdentificationThread(this, feat, size, database, options.identification.useG, options.identification.databasePercent, options.identification.matchAllDatabase);
	UpdateIdentificationProgress(FALSE, 0);
	UpdateTask();
	identification_time_started = GetTickCountEx();
	identification_thread->Resume();
	return true;
}

bool CMainForm::OnIdentificationProgress(bool started, IdenProgress *progress)
{
	if(stop_identification) return false;
	UpdateIdentificationProgress(started, started ? progress->count : (int)progress);
	if(started)
	{
		count_iden += progress->count;
		listview_results->SetRedrawState(false);
		for(int i = 0; i < progress->count; i++)
		{
			if(options.identification.showAllResults || VFSucceeded(progress->result[i]))
			{
				int j = listview_results->AddItem(IntToString(VFFeatGetG(database->Features(progress->index[i]))));
				listview_results->SetSubItemText(j, 1, VFResultToString(progress->result[i]));
				listview_results->SetSubItemText(j, 2, IntToString(progress->similarity[i]));
				listview_results->SetSubItemText(j, 3, IntToString(database->ID(progress->index[i])));
				listview_results->SetSubItemText(j, 4, database->FingerID(progress->index[i]));
			}
			if(!finger_found)
			{
				if(finger_id == database->FingerID(progress->index[i]))
				{
					finger_found = true;
				}
			}
		}
		listview_results->SetRedrawState(true);
	}
	return true;
}

void CMainForm::OnIdentificationComplete()
{
	double time = GetTickCountEx() - identification_time_started;
	INT result = identification_thread->Result();
	if(VFSucceeded(result) || result == VFE_FAILED)
	{
		main_log->Write(TEXT("Identification %s: %s matched"), stop_identification ? TEXT("stopped") : TEXT("complete"),
			(LPCTSTR)FingerprintCountToString(count_iden, STRING("no")));
		if(options.identification.useG && database->Count)
		{
			double p = PercentF(count_iden, database->Count);
			if(options.identification.matchAllDatabase && p == 100.0) main_log->Write(TEXT(" (all database)"));
			else main_log->Write(TEXT(" (%d%% of database size)"), Round(p));
		}
		main_log->WriteLine(TEXT(". Time: %s"), (LPCTSTR)TimeToString(time));
		if(finger_found && FileListRunning()) fingers_found++;
		int index = identification_thread->Index();
		int similarity = identification_thread->Similarity();
		if(index == -1)
		{
			main_log->Write(TEXT("Not identified"));
			if(finger_found && !finger_id.IsEmpty())
			{
				main_log->WriteLine(TEXT(". ID found in the database. False Rejection?"));
				if(FileListRunning()) false_rejected++;
			}
			else main_log->NewLine();
		}
		else
		{
			main_log->Write(TEXT("Identified finger ID: \"%s\""), (LPCTSTR)database->FingerID(index));
			if(options.identification.useG)
			{
				main_log->Write(TEXT(". G: %d"), VFFeatGetG(database->Features(index)));
			}
			main_log->Write(TEXT(". Similarity: %d"), similarity);
			string database_finger_id = database->FingerID(index);
			if(!finger_id.IsEmpty() && !database_finger_id.IsEmpty() && finger_id != database_finger_id)
			{
				main_log->WriteLine(TEXT(". IDs do not match. False Acceptance?"));
				if(FileListRunning()) false_accepted++;
			}
			else main_log->NewLine();
		}
		main_log->WriteLine(TEXT("Matching time: %s"), (LPCTSTR)TimeToString(identification_thread->Time()));
		if(count_iden)
		{
			main_log->WriteLine(TEXT("Matching speed: %s"), (LPCTSTR)MatchingSpeedToString(count_iden, identification_thread->Time()));
			if(options.identification.useG)
			{
				main_log->WriteLine(TEXT("Effective matching speed: %s"), (LPCTSTR)MatchingSpeedToString(database->Count, identification_thread->Time()));
			}
		}
	}
	else
	{
		main_log->WriteLine(TEXT("Identification error: %s"), (LPCTSTR)VFErrorToString(result));
		ErrorPrompt(Format(TEXT("Identification error: %s"), (LPCTSTR)VFErrorToString(result)));
	}
	if(FileListRunning())
	{
		total_iden += count_iden;
		total_iden_time += identification_thread->Time();
	}
	main_log->NewLine();
	delete identification_thread;
	identification_thread = NULL;
	stop_identification = false;
	UpdateIdentificationProgress(FALSE, 0);
	UpdateTask();
	if(FileListRunning()) task_thread->Resume();
}

void CMainForm::OnModeEnrollment()
{
	Mode = vfdmEnrollment;
}

void CMainForm::OnModeEnrollmentWithGen()
{
	Mode = vfdmEnrollmentWithGen;
}

void CMainForm::OnModeIdentification()
{
	Mode = vfdmIdentification;
}

void CMainForm::OnModeVerification()
{
	Mode = vfdmVerification;
}

void CMainForm::OnViewZoomIn()
{
	view_left->ZoomIn();
	view_right->ZoomIn();
	UpdateViewMenu();
}

void CMainForm::OnViewZoomOut()
{
	view_left->ZoomOut();
	view_right->ZoomOut();
	UpdateViewMenu();
}

void CMainForm::OnViewZoom1To1()
{
	view_left->Zoom = 100;
	view_right->Zoom = 100;
}

void CMainForm::OnViewShowFeatures()
{
	view_right->ShowFeatures = !view_right->ShowFeatures;
	UpdateViewMenu();
}

void CMainForm::OnViewShowSP()
{
	view_right->ShowSingularPoints = !view_right->ShowSingularPoints;
	UpdateViewMenu();
}

void CMainForm::OnViewShowBO()
{
	view_right->ShowBlockedOrientations = !view_right->ShowBlockedOrientations;
	UpdateViewMenu();
}

void CMainForm::UpdateViewMenu()
{
	EnableMenuItem(ID_VIEW_ZOOMIN, view_right->Zoom != CVFView::MaxZoom);
	EnableMenuItem(ID_VIEW_ZOOMOUT, view_right->Zoom != CVFView::MinZoom);
	CheckMenuItem(ID_VIEW_SHOWFEATURES, view_right->ShowFeatures, false);
	CheckMenuItem(ID_VIEW_SHOWSINGULARPOINTS, view_right->ShowSingularPoints, false);
	CheckMenuItem(ID_VIEW_SHOWBLOCKEDORIENTATIONS, view_right->ShowBlockedOrientations, false);
}

LRESULT CMainForm::OnSize(DWORD resizing_type, WORD client_width, WORD client_height)
{
	if(resizing_type == SIZE_MAXIMIZED || resizing_type == SIZE_RESTORED)
	{
		statusbar->SendMessage(WM_SIZE, 0, 0);
		AdjustControls(true, true);
		return 0;
	}
	else return CForm::OnSize(resizing_type, client_width, client_height);
}

void CMainForm::AdjustControls(bool top_part, bool bottom_part)
{
	if(top_part)
	{
		splitter_views->UpdateControls(splitter_horz->AreaTop(), splitter_horz->TopHeight(), view_left, view_right);
	}

	if(top_part && bottom_part) splitter_horz->UpdateControls(0, ClientWidth(), NULL, NULL);
	
	if(bottom_part)
	{
		int w;

		splitter_v1->UpdateControls(splitter_horz->Bottom(), splitter_horz->BottomHeight(), captionpanel_main, NULL);

		w = captionpanel_main->ClientWidth();
		panel_main->SetWidth(w);
		button_main_stop->SetRight(w - 8);
		label_main->SetWidth(button_main_stop->Left() - label_main->Left() - 8);
		progressbar_main->SetWidth(button_main_stop->Left() - progressbar_main->Left() - 8);
		label_main_time->SetWidth(button_main_stop->Left() - label_main_time->Left() - 8);
		richedit_main->SetSize(w, captionpanel_main->ClientHeight() - richedit_main->Top());

		splitter_v2->UpdateControls(splitter_horz->Bottom(), splitter_horz->BottomHeight(), captionpanel_scanner, captionpanel_results);

		richedit_scanner->SetSize(captionpanel_scanner->ClientWidth(), captionpanel_scanner->ClientHeight() - richedit_scanner->Top());

		w = captionpanel_results->ClientWidth();
		panel_results->SetWidth(w);
		button_results_stop->SetRight(w - 8);
		progressbar_results->SetWidth(button_results_stop->Left() - progressbar_results->Left() - 8);
		label_results_time->SetWidth(button_results_stop->Left() - label_results_time->Left() - 8);
		listview_results->SetSize(w, captionpanel_results->ClientHeight() - listview_results->Top());
	}
}

void CMainForm::CreateDB(const string &filename)
{
	HRSRC res = FindResource(NULL, MAKEINTRESOURCE(IDR_DATABASE), TEXT("DATABASE"));
	if(!res) throw exception("Resource not found");
	DWORD size = SizeofResource(NULL, res);
	HGLOBAL res_data = LoadResource(NULL, res);
	if(!res_data) throw exception("Can not load resource");
	LPVOID data = LockResource(res_data);
	if(!data) throw exception("Can not lock resource");
	HANDLE file = CreateFile(filename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if(file == INVALID_HANDLE_VALUE) throw exception("Can not create file");
	DWORD bytes_writen;
	bool error = !WriteFile(file, data, size, &bytes_writen, NULL) || bytes_writen != size;
	CloseHandle(file);
	if(error) throw exception("Error writing to file");
}

void CMainForm::OnToolsClearDatabase()
{
	if(!TaskRunning())
	{
		if(MessageBox(this, STRING("Do you really want to clear database?"), MB_YESNO | MB_ICONQUESTION) == IDYES)
		{
			CDBClearThread *thread = new CDBClearThread(NULL, database);
			DWORD tc = GetTickCount();
			WaitBox(this, thread, thread->TaskString());
			tc = GetTickCount() - tc;
			if(thread->Failed())
			{
				main_log->WriteLine(thread->TaskName() + TEXT(" failed"));
				ErrorBox(this, thread->Error());
			}
			else main_log->WriteLine(TEXT("Database cleared. Time: %s"), (LPCTSTR)TimeToString(tc));
			delete thread;
			main_log->NewLine();
			UpdateStatus();
		}
	}
}

void CMainForm::AddVersionToLog(CLog *log, const string &name, DWORD version_high, DWORD version_low)
{
	log->WriteLine(TEXT("%s    Version %u.%u.%u.%u"), (LPCTSTR)name,
		HIWORD(version_high), LOWORD(version_high), HIWORD(version_low), LOWORD(version_low));
}

void CMainForm::OnFirstShow()
{
	StartDBOpen();
}

void CMainForm::StartTask(CTaskThread *thread)
{
	task_thread = thread;
	UpdateTaskProgress(false, 0);
	UpdateTask();
	if(!main_log->LastLine().IsEmpty()) main_log->Write(TEXT(". "));
	main_log->WriteLine(task_thread->TaskName() + TEXT(" started..."));
	task_ui_time = 0;
	task_time_started = GetTickCount();
	task_thread->Resume();
}

void CMainForm::OnTaskComplete()
{
	DWORD time = GetTickCount() - task_time_started - task_ui_time;
	main_log->Write(task_thread->TaskName() + (task_thread->Failed() ? TEXT(" failed") : stop_task ? TEXT(" stopped") : TEXT(" complete")));
	if(DBOpenRunning()) OnDBOpenComplete(time);
	else if(FileListRunning()) OnFileListComplete(time);
	main_log->NewLine();
	delete task_thread;
	task_thread = NULL;
	stop_task = false;
	UpdateTaskProgress(false, 0);
	UpdateTask();
	ChangeBackMode();
}

void CMainForm::ChangeBackMode()
{
	if(changeBackMode)
	{
		Mode = lastMode;
		changeBackMode = false;
	}
}

void CMainForm::StartDBOpen()
{
	string db_name = GetFullFileName(databaseFileName);
	bool db_exists = false;
	if(GetFileName(db_name).IsEmpty() || !FileExists(db_name))
	{
		if(GetFileName(db_name).IsEmpty())
		{
			main_log->Write(TEXT("No database file specified"));
		}
		else
		{
			main_log->Write(TEXT("Database file not found"));
			if(createDatabase)
			{
				main_log->WriteLine(TEXT(". Creating"));
				try
				{
					CreateDB(db_name);
					db_exists = true;
				}
				catch(exception &e)
				{
					main_log->WriteLine(TEXT("Can not create database"));
					ErrorBox(this, e);
				}
			}
		}
	}
	else db_exists = true;
	if(db_exists)
	{
		if(!main_log->LastLine().IsEmpty()) main_log->NewLine();
		StartTask(new CDBOpenThread(this, database, defaultDatabaseProvider, db_name));
	}
	else
	{
		main_log->WriteLine(TEXT(". Working offline"));
		main_log->NewLine();
		ChangeBackMode();
	}
}

bool CMainForm::OnCloseQuery()
{
	return StopTaskPrompt();
}

void CMainForm::OnClose()
{
	CloseDB();
}

void CMainForm::OnToolsFileListEditor()
{
	FileListBox(this);
}

void CMainForm::OnToolsOptions()
{
	if(!TaskRunning())
	{
		if(OptionsBox(this, options))
		{
		}
	}
}

void CMainForm::OnToolsEngineOptions()
{
	if(!TaskRunning())
	{
		if(VFOptionsBox(this, vfOptions, engineName + " Options"))
		{
			VFUpdateOptions(vfOptions);
		}
	}
}

void CMainForm::OnHelpAbout()
{
	AboutBox(this, engineName);
}

string CMainForm::GetModeString()
{
	return VFDemoModeToString(mode);
}

string CMainForm::GetTaskName()
{
	if(task_thread) return task_thread->TaskName();
	else if(extraction_thread) return TEXT("Extraction");
	else if(identification_thread) return TEXT("Identification");
	else return emptyString;
}

string CMainForm::GetTaskString()
{
	if(task_thread) return task_thread->TaskString();
	else if(extraction_thread) return TEXT("Extracting");
	else if(identification_thread) return TEXT("Identificating");
	else if(mode == vfdmEnrollmentWithGen) return gen_index > 0 ? Format(TEXT("Waiting for image %d of %d"), gen_index + 1, VF_GENERALIZE_COUNT) : Format(TEXT("Waiting for %d images"), VF_GENERALIZE_COUNT);
	else if(mode == vfdmVerification) return ver_index > 0 ? Format(TEXT("Waiting for image %d of %d"), ver_index + 1, 2) : Format(TEXT("Waiting for %d images"), 2);
	else return TEXT("Waiting for image(s)");
}

void CMainForm::UpdateMode()
{
	CheckMenuItem(ID_MODE_ENROLLMENT, mode == vfdmEnrollment, true);
	CheckMenuItem(ID_MODE_ENROLLMENTWITHGEN, mode == vfdmEnrollmentWithGen, true);
	CheckMenuItem(ID_MODE_IDENTIFICATION, mode == vfdmIdentification, true);
	CheckMenuItem(ID_MODE_VERIFICATION, mode == vfdmVerification, true);
	UpdateTask();
}

void CMainForm::UpdateStatus()
{
	statusbar->SetPartText(0, TEXT("Records in the database: ") + IntToString(database->Count)
		+ (database->Opened() ? TEXT("") : TEXT(" (Offline)")));
	statusbar->SetPartText(1, database->DataSource());
}

void CMainForm::UpdateTask()
{
	bool running = TaskRunning();
	bool can_stop = CanStopTask();
	bool stopping = StoppingTask();
	EnableMenuItem(ID_FILE_CLEAR, !running);
	EnableMenuItem(ID_FILE_OPEN, !running && mode != vfdmNone);
	EnableMenuItem(ID_FILE_OPENFILELIST, !running && (mode == vfdmEnrollment || mode == vfdmIdentification));
	EnableMenuItem(ID_MODE_ENROLLMENT, !running);
	EnableMenuItem(ID_MODE_ENROLLMENTWITHGEN, !running);
	EnableMenuItem(ID_MODE_IDENTIFICATION, !running);
	EnableMenuItem(ID_MODE_VERIFICATION, !running);
	EnableMenuItem(ID_TOOLS_CLEARDATABASE, !running);
	EnableMenuItem(ID_TOOLS_CLEARMAINLOG, !running);
	EnableMenuItem(ID_TOOLS_CLEARSCANNERLOG, !running);
	EnableMenuItem(ID_TOOLS_OPTIONS, !running);
	EnableMenuItem(ID_TOOLS_ENGINEOPTIONS, !running);
	bool f;
	if(running) f = false;
	else
	{
		INT rt = VFRegistrationType();
		f = rt != VF_RT_NOT_PROTECTED && rt != VF_RT_HASP;
	}
	EnableMenuItem(ID_FILE_STOP, running && can_stop && !stopping);
	SetMenuItemText(ID_FILE_STOP, running ? (stopping ? TEXT("Stopping ") : TEXT("S&top ")) + GetTaskName() : TEXT("Stop"));

	string s = GetTaskString();
	if(running)
	{
		if(stopping) s += TEXT(" [stopping]");
	}
	SetText(Application()->Name() + (mode != vfdmNone ? STRING(" - ") + GetModeString() : emptyString) + (running ? TEXT(" - ") + s : emptyString));
	label_main->SetText(s);
	button_main_stop->SetEnabled(can_stop && !stopping);
	label_main_time->SetVisible(MainTaskRunning());

	button_results_stop->SetEnabled(IdentificationRunning() && !StoppingIdentification());
	label_results_time->SetVisible(IdentificationRunning());

	UpdateImagesMenu();
}

void CMainForm::UpdateImagesMenu()
{
	EnableMenuItem(ID_FILE_SAVE, !TaskRunning() && !image->IsEmpty());
	EnableMenuItem(ID_FILE_SAVERIGHT, !TaskRunning() && !image_after->IsEmpty());
}

void CMainForm::UpdateTaskProgress(bool started, int value, const string &text)
{
	static int progress, total;
	static DWORD time_started;
	if(started)
	{
		progress = value;
		progressbar_main->SetPosition(progress);
		DWORD time_elapsed = GetTickCount() - time_started - task_ui_time;
		DWORD time_estimated = progress && total != 0 ? Round((double)time_elapsed / progress * total) : 0;
		if(total != 0)
		{
			label_main_time->SetText(Format(TEXT("Time elapsed: %s"), (LPCTSTR)TimeToString(time_elapsed))
				+ (progress ? Format(TEXT("; remaining: %s"), (LPCTSTR)TimeToString(time_estimated - time_elapsed)) : emptyString));
		}
		else label_results_time->SetText(emptyString);
	}
	else
	{
		total = value;
		progress = 0;
		progressbar_main->SetMax(total);
		progressbar_main->SetVisible(total != 0);
		label_main_time->SetText(text);
		time_started = GetTickCount();
	}
}

void CMainForm::UpdateIdentificationProgress(bool started, int value)
{
	static int progress, total;
	static double time_started;
	if(started)
	{
		progress += value;
		progressbar_results->SetPosition(progress);
		double time_elapsed = GetTickCountEx() - time_started;
		double time_estimated = progress && total != 0 ? time_elapsed / progress * total : 0;
		if(total != 0)
		{
			label_results_time->SetText(Format(TEXT("Time elapsed: %s"), (LPCTSTR)TimeToString(time_elapsed))
				+ (progress ? Format(TEXT("; remaining: %s"), (LPCTSTR)TimeToString(time_estimated - time_elapsed)) : emptyString));
		}
		else label_results_time->SetText(emptyString);
	}
	else
	{
		total = value;
		progress = 0;
		progressbar_results->SetMax(total);
		progressbar_results->SetVisible(total != 0);
		label_results_time->SetText(emptyString);
		time_started = GetTickCountEx();
	}
}

bool CMainForm::TaskRunning()
{
	return task_thread != NULL || extraction_thread != NULL || identification_thread != NULL;
}

bool CMainForm::StoppingTask()
{
	return (task_thread && stop_task) || (!task_thread && identification_thread && stop_identification);
}

bool CMainForm::CanStopTask()
{
	return task_thread || identification_thread;
}

bool CMainForm::MainTaskRunning()
{
	return task_thread != NULL;
}

bool CMainForm::DBOpenRunning()
{
	return dynamic_cast<CDBOpenThread*>(task_thread) != 0;
}

bool CMainForm::FileListRunning()
{
	return dynamic_cast<CFileListThread*>(task_thread) != 0;
}

bool CMainForm::IdentificationRunning()
{
	return identification_thread != NULL;
}

bool CMainForm::StoppingIdentification()
{
	return stop_identification;
}

void CMainForm::Clear()
{
	if(!TaskRunning())
	{
		view_left->Image = NULL;
		view_right->Image = NULL;
		view_right->Features = NULL;
		image->Clear();
		image_after->Clear();
		features->Clear();
		finger_id.Clear();
		int i;
		for(i = 0; i < VF_GENERALIZE_COUNT; i++)
		{
			gen_image[i]->Clear();
			gen_image_after[i]->Clear();
			gen_features[i]->Clear();
			gen_finger_id[i].Clear();
		}
		gen_index = 0;
		for(i = 0; i < 2; i++)
		{
			ver_image[i]->Clear();
			ver_image_after[i]->Clear();
			ver_features[i]->Clear();
			ver_finger_id[i].Clear();
		}
		ver_index = 0;
		listview_results->ClearItems();
		UpdateTask();
	}
}

void CMainForm::OnToolsClearMainLog()
{
	if(!TaskRunning())
	{
		main_log->Clear();
	}
}

void CMainForm::OnToolsClearScannerLog()
{
	if(!TaskRunning())
	{
		scanner_log->Clear();
	}
}

void CMainForm::StopTask()
{
	if(task_thread)
	{
		stop_task = true;
		UpdateTask();
	}
	StopIdentification();
}

void CMainForm::StopIdentification()
{
	if(identification_thread && !stop_identification)
	{
		stop_identification = true;
		UpdateTask();
	}
}

bool CMainForm::StopTaskPrompt()
{
	if(TaskRunning() && CanStopTask() && !StoppingTask())
	{
		if(MessageBox(this, GetTaskString() + TEXT(" is in progress.\nDo you want to stop it?"),
			MB_YESNO | MB_ICONQUESTION) == IDYES) StopTask();
		return false;
	}
	return !TaskRunning();
}

bool CMainForm::ErrorPrompt(const string &reason, bool warning)
{
	DWORD error_time = GetTickCount();
	string text = reason;
	DWORD flags = warning ? MB_ICONWARNING : MB_ICONERROR;
	if(FileListRunning() && (mode == vfdmEnrollment || mode == vfdmIdentification))
	{
		text += TEXT("\nPress Cancel to stop file list");
		flags |= warning ? MB_YESNOCANCEL : MB_OKCANCEL;
	}
	else flags |= warning ? MB_YESNO : MB_OK;
	int result = MessageBox(this, text, flags);
	task_ui_time += GetTickCount() - error_time;
	switch(result)
	{
		case IDOK: case IDYES: return true;
		case IDNO: return false;
		default: StopTask(); return false;
	}
}
