using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Neurotec.Biometrics.VeriFingerDemo.Properties;

namespace Neurotec.Biometrics.VeriFingerDemo
{
	public partial class IdsForm : Form
	{
		#region Private fields

		private IDMacro macro;
		private IList<string> fileNames;
		private List<string> ids;

		#endregion

		#region Public constructor

		public IdsForm()
		{
			InitializeComponent();

			macro = new IDMacro();
			OnFileNamesChanged();
		}

		#endregion

		#region Private methods

		private void OnFileNamesChanged()
		{
			idsListView.BeginUpdate();
			idsListView.Items.Clear();
			if (fileNames != null)
			{
				foreach (string fileName in fileNames)
				{
					ListViewItem item = new ListViewItem(new string[] { Path.GetFileName(fileName), null });
					item.Tag = fileName;
					idsListView.Items.Add(item);
				}
			}
			idsListView.EndUpdate();
			UpdateIds();
		}

		private void UpdateIds()
		{
			if (fileNames != null)
			{
				try
				{
					macro.Code = ID;
					ids = new List<string>(fileNames.Count);
					foreach (string fileName in fileNames)
					{
						macro.Variables["Path"].Value = fileName;
						string id = macro.Execute().Trim();
						if (id == string.Empty) throw new MacroException("ID is empty", 0);
						ids.Add(id);
					}
					idErrorProvider.SetError(idTextBox, null);
				}
				catch (MacroException me)
				{
					ids = null;
					StringBuilder message = new StringBuilder();
					Exception e = me;
					int index = -1;
					while (e != null)
					{
						MacroException ime = e as MacroException;
						if (ime != null && ime.Index != -1) index = ime.Index;
						if (message.Length != 0) message.Append(Environment.NewLine);
						message.Append(e.Message);
						e = e.InnerException;
					}
					idErrorProvider.SetError(idTextBox, message.ToString());
					if (!idTextBox.Focused)
					{
						idTextBox.Select();
						idTextBox.Select(index, 0);
					}
				}
			}
			else
			{
				ids = null;
			}
			idsListView.BeginUpdate();
			int i = 0;
			foreach (ListViewItem item in idsListView.Items)
			{
				item.SubItems[1].Text = ids == null ? null : ids[i++];
			}
			idsListView.EndUpdate();
			okButton.Enabled = ids != null;
		}

		#endregion

		#region Public properties

		public IList<string> FileNames
		{
			get
			{
				return fileNames;
			}
			set
			{
				if (fileNames != value)
				{
					fileNames = value;
					OnFileNamesChanged();
				}
			}
		}

		public string ID
		{
			get
			{
				return idTextBox.Text;
			}
		}

		public IList<string> Ids
		{
			get
			{
				return ids;
			}
		}

		#endregion

		private void idTextBox_TextChanged(object sender, EventArgs e)
		{
			UpdateIds();
		}

		private void helpButton_Click(object sender, EventArgs e)
		{
			string help =
				"When enrolling fingerprints, id is stored in the database along with template extracted from the fingerprint image.\n"
				+ "Fingerprints from different fingers must have different ids. Same finger fingerprints must have same ids.\n"
				+ "When identifying a fingerprint in the database you will get finger id as a result.\n"
				+ "You can specify id directly for each fingerprint, like \"John's thumb\" or use macros in batch enroll mode to make id from fingerprint image file name.\n"
				+ "ID can contain any combination of characters and macros.\n"
				+ "Macros have to be in form of \"$(s)\", \"s\" can be one of the following functions:\n\n"

				+ "Path - returns path to the file\n"
				+ "FileName(s) - returns file name of s\n"
				+ "FileNameNoExt(s) - returns file name without extension of s\n"
				+ "Ext(s) - returns file extension of s\n"
				+ "DirectoryName(s) - returns file directory name of s\n"
				+ "PathNoExt(s) - returns file path without extension of s\n"
				+ "PathRoot(s) - returns path root of s\n"
				+ "SubdirName(s, i) - if i = 0 returns all directories that the path s contains\n"
					+ "\tif i > 0 returns i directory from the left of the path s\n"
					+ "\tif i < 0 returns -i directory from the right of the path s\n"
				+ "Left(s, n) - if n > 0, returns at most n characters from the left of s\n"
					+ "\tif n < 0, returns s without at most -n characters from the left\n"
					+ "\tif n = 0, returns empty string\n"
				+ "Right(s, n) - if n > 0, returns at most n characters from the right of s\n"
					+ "\tif n < 0, returns s without at most -n characters from the right\n"
					+ "\tif n = 0, returns empty string\n"
				+ "Center(s, i, n) - if n > 0, returns at most n characters from the position\n"
					+ "\tif n < 0, returns s without at most -n characters from the position\n"
					+ "\tif n = 0, returns empty string\n"
					+ "\tposition is i from the left if i > 0 and -i from the right if i < 0\n"
					+ "\tif i = 0 returns s\n"
				+ "Number(s, i) - if i = 0 returns empty string\n"
					+ "\tif i > 0, returns i-th number from the left of the string\n"
					+ "\tif i < 0, returns (-i)-th number from the right of the string\n"
				+ "NoNumber(s, i) - if i = 0 returns s\n"
					+ "\tif i > 0, returns s without i-th number from the left\n"
					+ "\tif i < 0, returns s without (-i)-th number from the right\n"

				+ "\n\ni and n parameters of the functions have to be an integer values.\n\n"
				+ "For example, if you have fingerprint images in files that are named \"xxxxy.bmp\", where \"xxxx\" is finger number and \"y\" is scanning number,\n"
				+ "then you can specify \"$(Right(FileNameNoExt(FileName),-1))\" macro to get \"xxxx\" as id from the file name\n";
			MessageBox.Show(help, Text);
		}

		private void IdsForm_FormClosed(object sender, FormClosedEventArgs e)
		{
			if (DialogResult == DialogResult.OK)
			{
				Settings.Default.Save();
			}
		}
	}

	public class IDMacro : Macro
	{
		#region Public constructor

		public IDMacro()
		{
			Variables.Add(new MacroVariable("Path", typeof(string)));
			MacroFunction leftFunction = new MacroFunction("Left", typeof(StringUtils).GetMethod("GetLeft"));
			MacroFunction rightFunction = new MacroFunction("Right", typeof(StringUtils).GetMethod("GetRight"));
			MacroFunction centerFunction = new MacroFunction("Center", typeof(StringUtils).GetMethod("GetCenter"));
			MacroFunction numberFunction = new MacroFunction("Number", typeof(StringUtils).GetMethod("GetNumber"));
			MacroFunction noNumberFunction = new MacroFunction("NoNumber", typeof(StringUtils).GetMethod("RemoveNumber"));
			MacroFunction fileNameFunction = new MacroFunction("FileName", typeof(Path).GetMethod("GetFileName"));
			MacroFunction fileNameNoExtFunction = new MacroFunction("FileNameNoExt", typeof(Path).GetMethod("GetFileNameWithoutExtension"));
			MacroFunction extFunction = new MacroFunction("Ext", typeof(Path).GetMethod("GetExtension"));
			MacroFunction directoryNameFunction = new MacroFunction("DirectoryName", typeof(Path).GetMethod("GetDirectoryName"));
			MacroFunction pathNoExtFunction = new MacroFunction("PathNoExt", typeof(PathUtils).GetMethod("GetPathWithoutExtension"));
			MacroFunction pathRootFunction = new MacroFunction("PathRoot", typeof(Path).GetMethod("GetPathRoot"));
			MacroFunction subdirNameFunction = new MacroFunction("SubdirName", typeof(PathUtils).GetMethod("GetSubdirectoryName"));
			Functions.Add(leftFunction);
			Functions.Add(rightFunction);
			Functions.Add(centerFunction);
			Functions.Add(numberFunction);
			Functions.Add(noNumberFunction);
			Functions.Add(fileNameFunction);
			Functions.Add(fileNameNoExtFunction);
			Functions.Add(extFunction);
			Functions.Add(directoryNameFunction);
			Functions.Add(pathNoExtFunction);
			Functions.Add(pathRootFunction);
			Functions.Add(subdirNameFunction);
		}

		#endregion
	}
}
