﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Collections;

namespace Neurotec.Biometrics.VeriFingerDemo
{
    public class MacroException: Exception
    {
        public MacroException(string message, int index)
            : this(message, index, null)
        {
        }

        public MacroException(string message, int index, Exception innerException)
            : base(message, innerException)
        {
            this.index = index;
        }

        private int index;

        public int Index
        {
            get
            {
                return index;
            }
        }
    }

    public abstract class MacroItem
    {
        public MacroItem(string name)
        {
            if(name == null) throw new ArgumentNullException("name");
            name = name.Trim();
            if(name == string.Empty) throw new ArgumentException("Name can not be empty string", "name");
            int index = 0;
            foreach(char c in name)
            {
                if((!char.IsLetterOrDigit(c) && c != '_') || (char.IsDigit(c) && index == 0))
                    throw new ArgumentException("Name must contain only letters, digits and underscore character. First character must not be a digit", "name:");
                index++;
            }
            if(name == "true" || name == "false" || name == "null")
                throw new ArgumentException("Name can not be \"true\", \"false\" or \"null\"", "name");
            this.name = name;
        }

        private string name;

        public string Name
        {
            get
            {
                return name;
            }
        }

        public abstract Type ItemType
        {
            get;
        }

        public abstract Type[] GetParameterTypes();
        public abstract object Execute(object[] parameters);
    }

    public class MacroVariable: MacroItem
    {
        public MacroVariable(string name, Type type)
            : base(name)
        {
            if(type == null) throw new ArgumentNullException("type");
            this.type = type;
            this.value = null;
        }

        private Type type;

        private object value;

        public object Value
        {
            get
            {
                return value;
            }
            set
            {
                if(value != null && value.GetType() != type)
                    throw new ArgumentException("Value is of wrong type");
                this.value = value;
            }
        }

        public override Type ItemType
        {
            get
            {
                return type;
            }
        }

        public override Type[] GetParameterTypes()
        {
            return Type.EmptyTypes;
        }

        public override object Execute(object[] parameters)
        {
            if(parameters == null) throw new ArgumentNullException();
            if(parameters.Length != 0) throw new ArgumentException("Variable do not take any parameters");
            if(value == null && type.IsValueType) throw new MacroException("Variable value is not set", -1);
            return value;
        }
    }

    public class MacroFunction: MacroItem
    {
        public MacroFunction(string name, MethodInfo methodInfo)
            : base(name)
        {
            if(methodInfo == null) throw new ArgumentNullException("methodInfo");
            if(!methodInfo.IsStatic || methodInfo.IsConstructor || !methodInfo.IsPublic
                || methodInfo.IsGenericMethodDefinition || methodInfo.ContainsGenericParameters || methodInfo.ReturnType == null)
                throw new ArgumentException("Method info must represent public static non-generic method with non-generic parameters and return value", "methodInfo");
            ParameterInfo[] parameterInfos = methodInfo.GetParameters();
            parameterTypes = new Type[parameterInfos.Length];
            int index = 0;
            foreach(ParameterInfo parameterInfo in parameterInfos)
            {
                if(parameterInfo.ParameterType.IsByRef || parameterInfo.IsOut || parameterInfo.IsRetval || parameterInfo.IsOptional)
                    throw new ArgumentException("Method info must represent method with by-value passed, in-only non-optional parameters", "methodInfo");
                parameterTypes[index++] = parameterInfo.ParameterType;
            }
            this.methodInfo = methodInfo;
        }

        private MethodInfo methodInfo;
        private Type[] parameterTypes;

        public override Type ItemType
        {
            get
            {
                return methodInfo.ReturnType;
            }
        }

        public override Type[] GetParameterTypes()
        {
            return parameterTypes;
        }

        public override object Execute(object[] parameters)
        {
            return methodInfo.Invoke(null, parameters);
        }
    }

    public class Macro
    {
        public Macro()
        {
            code = string.Empty;
            items = new MacroItemCollection();
            variables = new MacroVariableCollection(this);
            functions = new MacroFunctionCollection(this);
        }

        private string code;

        public string Code
        {
            get
            {
                return code;
            }
            set
            {
                if(value == null) throw new ArgumentNullException();
                code = value;
            }
        }

        private enum MacroStatus { Uninitialized, Waiting, WaitingSeparator, InName, InNumber, InString, InChar, Finished }

        private static void ThrowUnexpectedCharacter(char value, int index)
        {
            throw new MacroException(string.Format("Unexpected character '{0}'", value), index);
        }

        private object[] GetArguments(ref int index, Type[] paramTypes)
        {
            if(paramTypes == null) throw new ArgumentNullException("paramTypes");
            if(index < 0) throw new ArgumentOutOfRangeException("index", index, "Code index is less than zero");
            int length = code.Length;
            if(index > length) throw new ArgumentOutOfRangeException("index", index, "Code index is greater than length");
            int paramCount = paramTypes.Length;
            object[] values = new object[paramCount];
            int paramIndex = -1;
            Type type = null;
            bool strongType = false;
            object value = null;
            MacroStatus status = MacroStatus.Uninitialized;
            int startIndex = index;
            bool wasDigit = false;
            bool wasPeriod = false;
            bool wasExponent = false;
            bool wasExponentSign = false;
            bool wasExponentDigit = false;
            bool wasF = false;
            for(; index < length && status != MacroStatus.Finished; index++)
            {
                char c = code[index];
                bool isWhiteSpace = char.IsWhiteSpace(c);
                bool isComma = c == ',';
                bool isPeriod = c == '.';
                bool isOpening = c == '(';
                bool isClosing = c == ')';
                bool isSep = isComma || isOpening || isClosing;
                bool isSeparator = isWhiteSpace || isSep;
                bool isLetter = char.IsLetter(c);
                bool isDigit = char.IsDigit(c);
                bool isLetterOrDigit = isLetter || isDigit;
                bool isSign = c == '-' || c == '+';
                bool isDigitOrSign = isDigit || isSign;
                bool isUnderscore = c == '_';
                bool isLetterOrUnderscore = isLetter || isUnderscore;
                bool isLetterOrDigitOrUnderscore = isLetterOrDigit || isUnderscore;
                bool isDoubleQuote = c == '"';
                bool isQuote = c == '\'';
                bool isExponent = c == 'e' || c == 'E';
                bool isF = c == 'f';
                switch(status)
                {
                    case MacroStatus.Uninitialized:
                        if(!isWhiteSpace)
                        {
                            if(isOpening)
                            {
                                status = MacroStatus.Waiting;
                            }
                            else ThrowUnexpectedCharacter(c, index);
                        }
                        break;
                    case MacroStatus.Waiting:
                        if(isClosing)
                        {
                            status = MacroStatus.WaitingSeparator;
                            index--;
                        }
                        else if(!isWhiteSpace)
                        {
                            if(paramCount == 0)
                                throw new MacroException("No parameters expected", index);
                            if(isLetterOrUnderscore)
                            {
                                status = MacroStatus.InName;
                            }
                            else if(isDigitOrSign)
                            {
                                status = MacroStatus.InNumber;
                                wasDigit = c != '-';
                                wasPeriod = false;
                                wasExponent = false;
                                wasExponentSign = false;
                                wasExponentDigit = false;
                                wasF = false;
                            }
                            else if(isDoubleQuote)
                            {
                                status = MacroStatus.InString;
                            }
                            else if(isQuote)
                            {
                                status = MacroStatus.InChar;
                            }
                            else ThrowUnexpectedCharacter(c, index);
                            paramIndex++;
                            type = null;
                            strongType = true;
                            startIndex = index;
                        }
                        break;
                    case MacroStatus.WaitingSeparator:
                        if(!isWhiteSpace)
                        {
                            if(isComma)
                            {
                                if(paramIndex + 1 == paramCount)
                                    throw new MacroException(string.Format("Too many parameters (expected {0})", paramCount), index);
                                status = MacroStatus.Waiting;
                            }
                            else if(isClosing)
                            {
                                if(paramIndex + 1 < paramCount)
                                    throw new MacroException(string.Format("Too few parameters (expected {0})", paramCount), index);
                                status = MacroStatus.Finished;
                            }
                            else ThrowUnexpectedCharacter(c, index);
                            if(status != MacroStatus.WaitingSeparator)
                            {
                                Type paramType = paramTypes[paramIndex];
                                bool ok = paramType.IsAssignableFrom(type);
                                if(!ok)
                                {
                                    if(!paramType.IsValueType)
                                    {
                                        if(type == typeof(object) && !strongType && value == null) { ok = true; }
                                    }
                                    else if(paramType == typeof(double))
                                    {
                                        if(type == typeof(byte)) { value = (double)(byte)value; ok = true; }
                                        else if(type == typeof(sbyte)) { value = (double)(sbyte)value; ok = true; }
                                        else if(type == typeof(ushort)) { value = (double)(ushort)value; ok = true; }
                                        else if(type == typeof(short)) { value = (double)(short)value; ok = true; }
                                        else if(type == typeof(uint)) { value = (double)(uint)value; ok = true; }
                                        else if(type == typeof(int)) { value = (double)(int)value; ok = true; }
                                        else if(type == typeof(ulong)) { value = (double)(ulong)value; ok = true; }
                                        else if(type == typeof(long)) { value = (double)(long)value; ok = true; }
                                        else if(type == typeof(float)) { value = (double)(float)value; ok = true; }
                                    }
                                    else if(paramType == typeof(float))
                                    {
                                        if(type == typeof(byte)) { value = (double)(byte)value; ok = true; }
                                        else if(type == typeof(sbyte)) { value = (double)(sbyte)value; ok = true; }
                                        else if(type == typeof(ushort)) { value = (double)(ushort)value; ok = true; }
                                        else if(type == typeof(short)) { value = (double)(short)value; ok = true; }
                                        else if(type == typeof(uint)) { value = (double)(uint)value; ok = true; }
                                        else if(type == typeof(int)) { value = (double)(int)value; ok = true; }
                                    }
                                    else if(paramType == typeof(ulong))
                                    {
                                        if(type == typeof(byte)) { value = (ulong)(byte)value; ok = true; }
                                        else if(type == typeof(ushort)) { value = (ulong)(ushort)value; ok = true; }
                                        else if(type == typeof(uint)) { value = (ulong)(uint)value; ok = true; }
                                        else if(type == typeof(int) && !strongType && (int)value >= 0) { value = (ulong)(int)value; ok = true; }
                                    }
                                    else if(paramType == typeof(long))
                                    {
                                        if(type == typeof(byte)) { value = (long)(byte)value; ok = true; }
                                        else if(type == typeof(sbyte)) { value = (long)(sbyte)value; ok = true; }
                                        else if(type == typeof(ushort)) { value = (long)(ushort)value; ok = true; }
                                        else if(type == typeof(short)) { value = (long)(short)value; ok = true; }
                                        else if(type == typeof(uint)) { value = (long)(uint)value; ok = true; }
                                        else if(type == typeof(int)) { value = (long)(int)value; ok = true; }
                                    }
                                    else if(paramType == typeof(uint))
                                    {
                                        if(type == typeof(byte)) { value = (uint)(byte)value; ok = true; }
                                        else if(type == typeof(ushort)) { value = (uint)(ushort)value; ok = true; }
                                        else if(type == typeof(int) && !strongType && (int)value >= 0) { value = (uint)(int)value; ok = true; }
                                    }
                                    else if(paramType == typeof(int))
                                    {
                                        if(type == typeof(byte)) { value = (int)(byte)value; ok = true; }
                                        else if(type == typeof(sbyte)) { value = (int)(sbyte)value; ok = true; }
                                        else if(type == typeof(ushort)) { value = (int)(ushort)value; ok = true; }
                                        else if(type == typeof(short)) { value = (int)(short)value; ok = true; }
                                    }
                                    else if(paramType == typeof(ushort))
                                    {
                                        if(type == typeof(byte)) { value = (ushort)(byte)value; ok = true; }
                                        else if(type == typeof(int) && !strongType && (int)value >= 0 && (int)value <= ushort.MaxValue) { value = (ushort)(int)value; ok = true; }
                                    }
                                    else if(paramType == typeof(short))
                                    {
                                        if(type == typeof(byte)) { value = (short)(byte)value; ok = true; }
                                        else if(type == typeof(sbyte)) { value = (short)(sbyte)value; ok = true; }
                                        else if(type == typeof(int) && !strongType && (int)value >= short.MinValue && (int)value <= short.MaxValue) { value = (short)(int)value; ok = true; }
                                    }
                                    else if(paramType == typeof(byte))
                                    {
                                        if(type == typeof(int) && !strongType && (int)value >= byte.MinValue && (int)value <= byte.MaxValue) { value = (byte)(int)value; ok = true; }
                                    }
                                    else if(paramType == typeof(sbyte))
                                    {
                                        if(type == typeof(int) && !strongType && (int)value >= sbyte.MinValue && (int)value <= sbyte.MaxValue) { value = (sbyte)(int)value; ok = true; }
                                    }
                                }
                                if(!ok) throw new MacroException(string.Format("Incompatible types for parameter {0}: excpected {1}, found {2}", paramIndex, paramType, type), startIndex);
                                values[paramIndex] = value;
                            }
                        }
                        break;
                    case MacroStatus.InName:
                        if(isSeparator)
                        {
                            status = MacroStatus.WaitingSeparator;
                            string name = code.Substring(startIndex, index - startIndex);
                            if(name == "true" || name == "false")
                            {
                                type = typeof(bool);
                                value = name == "true";
                            }
                            else if(name == "null")
                            {
                                type = typeof(object);
                                strongType = false;
                                value = null;
                            }
                            else
                            {
                                MacroItem item = items[name];
                                if(item == null)
                                    throw new MacroException(string.Format("Macro function or variable named \"{0}\" not found", name), startIndex);
                                object[] itemParams;
                                if(item is MacroVariable)
                                {
                                    itemParams = new object[0];
                                }
                                else
                                {
                                    try
                                    {
                                        itemParams = GetArguments(ref index, item.GetParameterTypes());
                                    }
                                    catch(MacroException e)
                                    {
                                        throw new MacroException(string.Format("Error parsing function \"{0}\" arguments", name), startIndex, e);
                                    }
                                }
                                type = item.ItemType;
                                try
                                {
                                    value = item.Execute(itemParams);
                                }
                                catch(Exception e)
                                {
                                    throw new MacroException(string.Format("Error executing function \"{0}\"", name), startIndex, e);
                                }
                            }
                            index--;
                        }
                        else if(!isLetterOrDigitOrUnderscore) ThrowUnexpectedCharacter(c, index);
                        break;
                    case MacroStatus.InNumber:
                        if((wasDigit && isWhiteSpace) || isSep)
                        {
                            if(wasExponent && !wasExponentDigit)
                                throw new MacroException("Exponent without digits in number", startIndex);
                            status = MacroStatus.WaitingSeparator;
                            string number = code.Substring(startIndex, index - startIndex - (wasF ? 1 : 0));
                            try
                            {
                                if(wasPeriod || wasExponent)
                                {
                                    if(wasF)
                                    {
                                        type = typeof(float);
                                        value = float.Parse(number);
                                    }
                                    else
                                    {
                                        type = typeof(double);
                                        value = double.Parse(number);
                                    }
                                }
                                else
                                {
                                    type = typeof(int);
                                    strongType = false;
                                    value = int.Parse(number);
                                }
                            }
                            catch(FormatException e)
                            {
                                throw new MacroException("Invalid number constant", startIndex, e);
                            }
                            catch(OverflowException e)
                            {
                                throw new MacroException("Too big number constant", startIndex, e);
                            }
                            index--;
                        }
                        else
                        {
                            if(wasF) ThrowUnexpectedCharacter(c, index);
                            if(isPeriod)
                            {
                                if(!wasDigit || wasPeriod) ThrowUnexpectedCharacter(c, index);
                                wasPeriod = true;
                            }
                            else if(isExponent)
                            {
                                if(!wasDigit || wasExponent) ThrowUnexpectedCharacter(c, index);
                                wasExponent = true;
                            }
                            else if(isSign)
                            {
                                if(!wasExponent || wasExponentSign || wasExponentDigit) ThrowUnexpectedCharacter(c, index);
                                wasExponentSign = true;
                            }
                            else if(isDigit)
                            {
                                wasDigit = true;
                                if(wasExponent) wasExponentDigit = true;
                            }
                            else if(isF)
                            {
                                if(wasF || !wasDigit || (wasExponent && !wasExponentDigit)) ThrowUnexpectedCharacter(c, index);
                                wasF = true;
                            }
                            else if(!isWhiteSpace) ThrowUnexpectedCharacter(c, index);
                        }
                        break;
                    case MacroStatus.InString:
                        if(isDoubleQuote)
                        {
                            status = MacroStatus.WaitingSeparator;
                            type = typeof(string);
                            value = code.Substring(startIndex + 1, index - startIndex - 1);
                        }
                        break;
                    case MacroStatus.InChar:
                        if(isQuote)
                        {
                            status = MacroStatus.WaitingSeparator;
                            try
                            {
                                type = typeof(char);
                                value = char.Parse(code.Substring(startIndex + 1, index - startIndex - 1));
                            }
                            catch(FormatException e)
                            {
                                throw new MacroException("Invalid char constant", startIndex, e);
                            }
                        }
                        break;
                }
            }
            if(status == MacroStatus.Uninitialized)
                throw new MacroException("Arguments not found", index);
            return values;
        }

        public string Execute()
        {
            StringBuilder value = new StringBuilder();
            int length = code.Length;
            for(int i = 0; i < length; i++)
            {
                char c = code[i];
                if(c == '$')
                {
                    if(i != length - 1 && code[i + 1] == c)
                    {
                        value.Append(c);
                        i++;
                    }
                    else
                    {
                        object[] arguments;
                        try
                        {
                            i++;
                            arguments = GetArguments(ref i, new Type[] { typeof(object) });
                        }
                        catch(MacroException e)
                        {
                            throw new MacroException("Error parsing macro", i, e);
                        }
                        value.Append(arguments[0]);
                        i--;
                    }
                }
                else if(c == '/')
                {
                    if(i != length - 1 && code[i + 1] == c)
                    {
                        value.Append(c);
                        i++;
                    }
                    else if(value.Length == 0 || value[value.Length - 1] != '\\') value.Append('\\');
                }
                else value.Append(c);
            }
            return value.ToString();
        }

        private MacroItemCollection items;
        private MacroVariableCollection variables;
        private MacroFunctionCollection functions;

        public MacroItemCollection Items
        {
            get
            {
                return items;
            }
        }

        public MacroVariableCollection Variables
        {
            get
            {
                return variables;
            }
        }

        public MacroFunctionCollection Functions
        {
            get
            {
                return functions;
            }
        }

        public class MacroItemCollection: ICollection<MacroItem>, IEnumerable<MacroItem>
        {
            internal MacroItemCollection()
            {
                items = new List<MacroItem>();
            }

            private List<MacroItem> items;

            #region ICollection<MacroItem> Members

            public void Add(MacroItem item)
            {
                if(item == null) throw new ArgumentNullException("item");
                int index = 0;
                foreach(MacroItem it in items)
                {
                    int c = it.Name.CompareTo(item.Name);
                    if(c == 0) throw new ArgumentException("Item with the same name already exists", "item");
                    if(c > 0) break;
                    index++;
                }
                items.Insert(index, item);
            }

            public void Clear()
            {
                items.Clear();
            }

            public bool Contains(MacroItem item)
            {
                return items.Contains(item);
            }

            public void CopyTo(MacroItem[] array, int arrayIndex)
            {
                items.CopyTo(array, arrayIndex);
            }

            public int Count
            {
                get
                {
                    return items.Count;
                }
            }

            bool ICollection<MacroItem>.IsReadOnly
            {
                get { return false; }
            }

            public bool Remove(MacroItem item)
            {
                return items.Remove(item);
            }

            #endregion

			#region IEnumerable Members

			IEnumerator IEnumerable.GetEnumerator()
			{
				return GetEnumerator();
			}

			#endregion

			#region IEnumerable<MacroItem> Members

            public IEnumerator<MacroItem> GetEnumerator()
            {
                return items.GetEnumerator();
            }

            #endregion

            public int IndexOf(MacroItem item)
            {
                return items.IndexOf(item);
            }

            public void RemoveAt(int index)
            {
                items.RemoveAt(index);
            }

            public MacroItem this[int index]
            {
                get
                {
                    return items[index];
                }
            }

            public MacroItem this[string name]
            {
                get
                {
                    foreach(MacroItem item in items)
                        if(item.Name == name) return item;
                    return null;
                }
            }
		}

        public class MacroVariableCollection: ICollection<MacroVariable>, IEnumerable<MacroVariable>
        {
            internal MacroVariableCollection(Macro macro)
            {
                if(macro == null) throw new ArgumentNullException("macro");
                this.macro = macro;
            }

            private Macro macro;

            #region ICollection<MacroVariable> Members

            public void Add(MacroVariable item)
            {
                macro.items.Add(item);
            }

            public void Clear()
            {
                for(int i = macro.items.Count - 1; i >= 0; i--)
                    if(macro.items[i] is MacroVariable) macro.items.RemoveAt(i);
            }

            public bool Contains(MacroVariable item)
            {
                return macro.items.Contains(item);
            }

            public void CopyTo(MacroVariable[] array, int arrayIndex)
            {
                if(array == null) throw new ArgumentNullException("array");
                if(arrayIndex < 0) throw new ArgumentException("Array index is less than zero", "arrayIndex");
                foreach(MacroItem item in macro.items)
                {
                    if(item is MacroVariable)
                    {
                        if(arrayIndex >= array.Length)
                            throw new ArgumentException("Array index plus number of items is greater than length of the array");
                        array[arrayIndex++] = (MacroVariable)item;
                    }
                }
            }

            public int Count
            {
                get
                {
                    int count = 0;
                    foreach(MacroItem item in macro.items)
                    {
                        if(item is MacroVariable)
                            count++;
                    }
                    return count;
                }
            }

            bool ICollection<MacroVariable>.IsReadOnly
            {
                get { return false; }
            }

            public bool Remove(MacroVariable item)
            {
                return macro.items.Remove(item);
            }

            #endregion

			#region IEnumerable Members

			IEnumerator IEnumerable.GetEnumerator()
			{
				return GetEnumerator();
			}

			#endregion

			#region IEnumerable<MacroVariable> Members

            public IEnumerator<MacroVariable> GetEnumerator()
            {
                return new MacroVariableCollectionEnumerator(macro.items.GetEnumerator());
            }

            private class MacroVariableCollectionEnumerator: IEnumerator<MacroVariable>
            {
                public MacroVariableCollectionEnumerator(IEnumerator<MacroItem> enumerator)
                {
                    if(enumerator == null) throw new ArgumentNullException("enumerator");
                    this.enumerator = enumerator;
                }

                private IEnumerator<MacroItem> enumerator;

				#region IEnumerator Members

				object IEnumerator.Current
				{
					get
					{
						return Current;
					}
				}

				public bool MoveNext()
				{
					while(true)
					{
						if(!enumerator.MoveNext()) return false;
						if(enumerator.Current is MacroVariable) return true;
					}
				}

				public void Reset()
				{
					enumerator.Reset();
				}

				#endregion

                #region IEnumerator<MacroVariable> Members

                public MacroVariable Current
                {
                    get
                    {
                        return (MacroVariable)enumerator.Current;
                    }
                }

                #endregion

                #region IDisposable Members

                public void Dispose()
                {
                }

                #endregion
			}

            #endregion

			public MacroVariable this[string name]
            {
                get
                {
                    foreach(MacroItem item in macro.items)
                    {
                        if(item is MacroVariable && item.Name == name) return (MacroVariable)item;
                    }
                    return null;
                }
            }
		}

        public class MacroFunctionCollection: ICollection<MacroFunction>, IEnumerable<MacroFunction>
        {
            internal MacroFunctionCollection(Macro macro)
            {
                if(macro == null) throw new ArgumentNullException("macro");
                this.macro = macro;
            }

            private Macro macro;

            #region ICollection<MacroFunction> Members

            public void Add(MacroFunction item)
            {
                macro.items.Add(item);
            }

            public void Clear()
            {
                for(int i = macro.items.Count - 1; i >= 0; i--)
                    if(macro.items[i] is MacroFunction) macro.items.RemoveAt(i);
            }

            public bool Contains(MacroFunction item)
            {
                return macro.items.Contains(item);
            }

            public void CopyTo(MacroFunction[] array, int arrayIndex)
            {
                if(array == null) throw new ArgumentNullException("array");
                if(arrayIndex < 0) throw new ArgumentException("Array index is less than zero", "arrayIndex");
                foreach(MacroItem item in macro.items)
                {
                    if(item is MacroFunction)
                    {
                        if(arrayIndex >= array.Length)
                            throw new ArgumentException("Array index plus number of items is greater than length of the array");
                        array[arrayIndex++] = (MacroFunction)item;
                    }
                }
            }

            public int Count
            {
                get
                {
                    int count = 0;
                    foreach(MacroItem item in macro.items)
                    {
                        if(item is MacroFunction)
                            count++;
                    }
                    return count;
                }
            }

            bool ICollection<MacroFunction>.IsReadOnly
            {
                get { return false; }
            }

            public bool Remove(MacroFunction item)
            {
                return macro.items.Remove(item);
            }

            #endregion

			#region IEnumerable Members

			IEnumerator IEnumerable.GetEnumerator()
			{
				return GetEnumerator();
			}

			#endregion

			#region IEnumerable<MacroFunction> Members

            public IEnumerator<MacroFunction> GetEnumerator()
            {
                return new MacroFunctionCollectionEnumerator(macro.items.GetEnumerator());
            }

            private class MacroFunctionCollectionEnumerator: IEnumerator<MacroFunction>
            {
                public MacroFunctionCollectionEnumerator(IEnumerator<MacroItem> enumerator)
                {
                    if(enumerator == null) throw new ArgumentNullException("enumerator");
                    this.enumerator = enumerator;
                }

                private IEnumerator<MacroItem> enumerator;

				#region IEnumerator Members

				object IEnumerator.Current
				{
					get
					{
						return Current;
					}
				}

				public bool MoveNext()
				{
					while(true)
					{
						if(!enumerator.MoveNext()) return false;
						if(enumerator.Current is MacroFunction) return true;
					}
				}

				public void Reset()
				{
					enumerator.Reset();
				}

				#endregion

                #region IEnumerator<MacroFunction> Members

                public MacroFunction Current
                {
                    get
                    {
                        return (MacroFunction)enumerator.Current;
                    }
                }

                #endregion

                #region IDisposable Members

                public void Dispose()
                {
                }

                #endregion
			}

            #endregion

            public MacroFunction this[string name]
            {
                get
                {
                    foreach(MacroItem item in macro.items)
                    {
                        if(item is MacroFunction && item.Name == name) return (MacroFunction)item;
                    }
                    return null;
                }
            }
		}
    }
}